//----------------------------------------------------------------------------------------
// basis.H: Contains declarations for basis functions on 2D and 3D elements
//----------------------------------------------------------------------------------------

#ifndef BASIS_H
#define BASIS_H

#include <iostream>
#include <fstream>
#include <cmath>
#include "general.H"
#include "map_info.H"

//----------------------------------------------------------------------------------------
// BASIS Class: Abstract base class: Defines common interface for all BASIS classes
//----------------------------------------------------------------------------------------

class BASIS
{
private:
  int dim;  // Number of degrees of freedom in the basis

protected:
  // Compute values of basis functions at a point
  // Parameters: coord - point coordinates; result - values of basis functions
  virtual void basis_func(const real coord[], func_type result) const
  {std::cerr << "File basis.H: Calling a basis class member function that was not defined!\n"; std::exit(1);}

  // Compute values of gradients of basis functions at a point
  // Parameters: coord - point coordinates; result - values of gradients of basis functions
  virtual void basis_grad(const real coord[], grad_type result) const
  {std::cerr << "File basis.H: Calling a basis class member function that was not defined!\n"; std::exit(1);}

  // Compute values of basis functions at a point
  // Parameters: coord - point coordinates; result - values of basis functions
  virtual void basis_func_vert_integral(const real coord[], func_type result) const
  {std::cerr << "File basis.H: Calling a basis class member function that was not defined!\n"; std::exit(1);}

public:
  // Default constructor
  BASIS() {}

  // Default destructor
  ~BASIS() {}

  // Return value: Number of degrees of freedom in the basis: rvalue
  int Dim() const {return dim;}

  // Return value: Number of degrees of freedom in the basis: lvalue
  int& Dim() {return dim;}

  // Compute values of basis functions at a point in physical space
  // Parameters: coord - point coordinates; mapping - map object from the reference element;
  //             result - values of basis functions
  virtual void basis_func_phys(const real coord[], const MAP& mapping, func_type result) const 
  {std::cerr << "File basis.H: Calling a basis class member function that was not defined!\n"; std::exit(1);}

  // Compute values of gradients of basis functions at a point in physical space
  // Parameters: coord - point coordinates; mapping - map object from the reference element;
  //             result - values of gradients of basis functions
  virtual void basis_grad_phys(const real coord[], const MAP& mapping, grad_type result) const
  {std::cerr << "File basis.H: Calling a basis class member function that was not defined!\n"; std::exit(1);}

  // Compute values of basis functions at a point in reference space
  // Parameters: coord - point coordinates; mapping - map object from the reference element;
  //             result - values of basis functions
  virtual void basis_func_ref(const real coord[], const MAP& mapping, func_type result) const 
  {std::cerr << "File basis.H: Calling a basis class member function that was not defined!\n"; std::exit(1);}

  // Compute values of gradients of basis functions at a point in reference space
  // Parameters: coord - point coordinates; mapping - map object from the reference element;
  //             result - values of gradients of basis functions
  virtual void basis_grad_ref(const real coord[], const MAP& mapping, grad_type result) const
  {std::cerr << "File basis.H: Calling a basis class member function that was not defined!\n"; std::exit(1);}

  // Compute values of basis functions at a point in reference space
  // Parameters: coord - point coordinates; result - values of basis functions
  virtual void basis_func_ref(const real coord[], func_type result) const 
  {std::cerr << "File basis.H: Calling a basis class member function that was not defined!\n"; std::exit(1);}

  // Compute values of gradients of basis functions at a point in reference space
  // Parameters: coord - point coordinates; result - values of gradients of basis functions
  virtual void basis_grad_ref(const real coord[], grad_type result) const
  {std::cerr << "File basis.H: Calling a basis class member function that was not defined!\n"; std::exit(1);}

  // Compute values of integrals in the z-direction of basis functions at a point
  // Parameters: coord - point coordinates; mapping - map object from the reference element;
  //             result - values of integrals in the z-direction of basis functions
  virtual void basis_func_vert_integral_phys(const real coord[], const MAP& mapping, func_type result) const
  {std::cerr << "File basis.H: Calling a basis class member function that was not defined!\n"; std::exit(1);}
};



//--------------------------------------------------------------------------------------
// BASIS_2D Class: Defines basis functions for 2D elements
//--------------------------------------------------------------------------------------

// Basis functions on the reference element in 2D
class BASIS_2D : public BASIS
{
public:
  // Default constructor
  BASIS_2D(){}

  // Destructor
  ~BASIS_2D(){}

  // Compute values of basis functions at a point in physical space
  // Parameters: x - point coordinates; mapping - map object from the reference element;
  //             result - values of basis functions
  void basis_func_phys(const real x[], const MAP& mapping, func_type result) const
  {real xi[d_space_2d]; mapping.mapto_inverse(x, xi); basis_func(xi, result);}

  // Compute values of gradients of basis functions at a point in physical space
  // Parameters: x - point coordinates; mapping - map object from the reference element;
  //             result - values of gradients of basis functions
  void basis_grad_phys(const real x[], const MAP& mapping, grad_type result) const
  {
    real xi[d_space_2d];
    grad_type grad_tmp;

    mapping.mapto_inverse(x, xi);
    basis_grad(xi, grad_tmp);
    for (int j=0; j<Dim(); j++)
      ((MapTriangle&)mapping).MapGradient(grad_tmp[j], result[j]);
  }

  // Compute values of basis functions at a point in reference space
  // Parameters: xi - point coordinates; mapping - map object from the reference element;
  //             result - values of basis functions
  void basis_func_ref(const real xi[], const MAP& mapping, func_type result) const
  {basis_func(xi, result);}

  // Compute values of gradients of basis functions at a point in reference space
  // Parameters: xi - point coordinates; mapping - map object from the reference element;
  //             result - values of gradients of basis functions
  void basis_grad_ref(const real xi[], const MAP& mapping, grad_type result) const
  {
    grad_type grad_tmp;

    basis_grad(xi, grad_tmp);
    for (int j=0; j<Dim(); j++)
      ((MapTriangle&)mapping).MapGradient(grad_tmp[j], result[j]);
  }

  // Compute values of basis functions at a point in reference space
  // Parameters: xi - point coordinates; result - values of basis functions
  void basis_func_ref(const real xi[], func_type result) const {basis_func(xi, result);}

  // Compute values of gradients of basis functions at a point in reference space
  // Parameters: xi - point coordinates; result - values of gradients of basis functions
  void basis_grad_ref(const real xi[], grad_type result) const {basis_grad(xi, result);}
};

// Basis functions on the reference element in 2D
class BASIS_2D_CONST : public BASIS_2D
{
protected:
  // Compute values of basis functions at a point
  // Parameters: xi - point coordinates; result - values of basis functions
  void basis_func(const real xi[], func_type result) const
  {
    static long_real sqrt2=std::sqrt(2.);

    result[0]=sqrt2;
  }

  // Compute values of gradients of basis functions at a point
  // Parameters: xi - point coordinates; result - values of gradients of basis functions
  void basis_grad(const real xi[], grad_type result) const
  {
    result[0][0]=result[0][1]=0.;
  }

public:
  // Default constructor
  BASIS_2D_CONST() {Dim()=1;}

  // Destructor
  ~BASIS_2D_CONST() {}
};

// Basis functions on the reference element in 2D
class BASIS_2D_LIN : public BASIS_2D
{
protected:
  // Compute values of basis functions at a point
  // Parameters: xi - point coordinates; result - values of basis functions
  void basis_func(const real xi[], func_type result) const
  {
    static long_real sqrt2=std::sqrt(2.), sqrt3=std::sqrt(3.);

    result[2]=2.*sqrt3*(1.-xi[0]-2.*xi[1]);
    result[1]=2.-6.*xi[0];
    result[0]=sqrt2;
  }

  // Compute values of gradients of basis functions at a point
  // Parameters: xi - point coordinates; result - values of gradients of basis functions
  void basis_grad(const real xi[], grad_type result) const
  {
    static long_real sqrt3=std::sqrt(3.);

    result[2][0]=-2.*sqrt3; result[2][1]=-4.*sqrt3;
    result[1][0]=-6.; result[1][1]=0.;
    result[0][0]=result[0][1]=0.;
  }

public:
  // Default constructor
  BASIS_2D_LIN() {Dim()=3;}

  // Destructor
  ~BASIS_2D_LIN() {}
};

// Basis functions on the reference element in 2D
class BASIS_2D_QUAD : public BASIS_2D
{
protected:
  // Compute values of basis functions at a point
  // Parameters: xi - point coordinates; result - values of basis functions
  void basis_func(const real xi[], func_type result) const
  {
    static long_real sqrt2=std::sqrt(2.), sqrt3=std::sqrt(3.),
      three_sqrt5=3.*std::sqrt(5.), sqrt6=std::sqrt(6.);

    result[5]=three_sqrt5*((3.*xi[0]+8.*xi[1]-4.)*xi[0]+(3.*xi[1]-4.)*xi[1]+1.);
    result[4]=sqrt3*((5.*xi[0]-4.)*xi[0]+(-15.*xi[1]+12.)*xi[1]-1.);
    result[3]=sqrt6*((10.*xi[0]-8.)*xi[0]+1.);
    result[2]=2.*sqrt3*(1.-xi[0]-2.*xi[1]);
    result[1]=2.-6.*xi[0];
    result[0]=sqrt2;
  }

  // Compute values of gradients of basis functions at a point
  // Parameters: xi - point coordinates; result - values of gradients of basis functions
  void basis_grad(const real xi[], grad_type result) const
  {
    static long_real sqrt3=std::sqrt(3.), three_sqrt5=3.*std::sqrt(5.), sqrt6=std::sqrt(6.);

    result[5][0]=three_sqrt5*(6.*xi[0]+8.*xi[1]-4.); 
    result[5][1]=three_sqrt5*(8.*xi[0]+6.*xi[1]-4.);
    result[4][0]=sqrt3*(10.*xi[0]-4.); result[4][1]=sqrt3*(-30.*xi[1]+12.);
    result[3][0]=sqrt6*(20.*xi[0]-8.); result[3][1]=0.;
    result[2][0]=-2.*sqrt3; result[2][1]=-4.*sqrt3;
    result[1][0]=-6.; result[1][1]=0.;
    result[0][0]=result[0][1]=0.;
  }
public:
  // Default constructor
  BASIS_2D_QUAD() {Dim()=6;}

  // Destructor
  ~BASIS_2D_QUAD() {}
};




//--------------------------------------------------------------------------------------
// BASIS_REF Class: Defines basis functions in reference space for 3D elements
//                  used in parallel prisms
//--------------------------------------------------------------------------------------

class BASIS_REF : public BASIS
{
public:
  // Default constructor
  BASIS_REF() {}

  // Destructor
  ~BASIS_REF() {}

  // Compute values of basis functions at a point in physical space
  // Parameters: x - point coordinates; mapping - map object from the reference element;
  //             result - values of basis functions
  void basis_func_phys(const real x[], const MAP& mapping, func_type result) const
  {real xi[d_space]; mapping.mapto_inverse(x, xi); basis_func(xi, result);}

  // Compute values of gradients of basis functions at a point in physical space
  // Parameters: x - point coordinates; mapping - map object from the reference element;
  //             result - values of gradients of basis functions
  void basis_grad_phys(const real x[], const MAP& mapping, grad_type result) const
  {
    real xi[d_space];
    grad_type grad_tmp;

    mapping.mapto_inverse(x, xi);
    basis_grad(xi, grad_tmp);
    for (int j=0; j<Dim(); j++)
      ((MapPrism&)mapping).MapGradient(grad_tmp[j], result[j]);
  }

  // Compute values of basis functions at a point in reference space
  // Parameters: xi - point coordinates; mapping - map object from the reference element;
  //             result - values of basis functions
  void basis_func_ref(const real xi[], const MAP& mapping, func_type result) const
  {basis_func(xi, result);}

  // Compute values of gradients of basis functions at a point in reference space
  // Parameters: xi - point coordinates; mapping - map object from the reference element;
  //             result - values of gradients of basis functions
  void basis_grad_ref(const real xi[], const MAP& mapping, grad_type result) const
  {
    grad_type grad_tmp;

    basis_grad(xi, grad_tmp);
    for (int j=0; j<Dim(); j++)
      ((MapPrism&)mapping).MapGradient(grad_tmp[j], result[j]);
  }

  // Compute values of basis functions at a point in reference space
  // Parameters: xi - point coordinates; result - values of basis functions
  void basis_func_ref(const real xi[], func_type result) const {basis_func(xi, result);}

  // Compute values of gradients of basis functions at a point in reference space
  // Parameters: xi - point coordinates; result - values of gradients of basis functions
  void basis_grad_ref(const real xi[], grad_type result) const {basis_grad(xi, result);}

  // Compute values of integrals in the z-direction of basis functions at a point
  // Parameters: x - point coordinates; mapping - map object from the reference element;
  //             result - values of integrals in the z-direction of basis functions
  void basis_func_vert_integral_phys(const real x[], const MAP& mapping, func_type result) const
  {
    real xi[d_space], height;
    func_type result_p;

    mapping.mapto_inverse(x, xi);
    basis_func_vert_integral(xi, result);
    height=((MapPrism&)mapping).GetHeight(xi);
    xi[2]=0.;
    basis_func_vert_integral(xi, result_p);

    for (int j=0; j<Dim(); j++)
      result[j]-=result_p[j], result[j]*=height;
  }

//   void basis_func_vert_integral_phys(const real x[], const MAP& mapping, func_type result) const
//   {
//     real xi[d_space];
//     func_type result_p;

//     mapping.mapto_inverse(x, xi);
//     basis_func_vert_integral(xi, result_p);
//     real height=((MapPrism&)mapping).GetHeight(xi);
//     xi[2]=1.;
//     basis_func_vert_integral(xi, result);

//     for (int j=0; j<Dim(); j++)
//       result[j]-=result_p[j], result[j]*=height;
//   }
};

// Basis functions on the reference element (used on parallel prisms)
class BASIS_REF_CONST : public BASIS_REF
{
protected:
  // Compute values of basis functions at a point
  // Parameters: xi - point coordinates; result - values of basis functions
  void basis_func(const real xi[], func_type result) const
  {
    static long_real sqrt2=std::sqrt(2.);

    result[0]=sqrt2;
  }

  // Compute values of gradients of basis functions at a point
  // Parameters: xi - point coordinates; result - values of gradients of basis functions
  void basis_grad(const real xi[], grad_type result) const
  {
    result[0][0]=result[0][1]=result[0][2]=0.;
  }

  // Compute values of integrals in the z-direction of basis functions at a point
  // Parameters: xi - point coordinates;
  //             result - values of integrals in the z-direction of basis functions
  void basis_func_vert_integral(const real xi[], func_type result) const
  {
    static long_real sqrt2=std::sqrt(2.);

    result[0]=sqrt2*xi[2];
  }

public:
  // Default constructor
  BASIS_REF_CONST() {Dim()=1;}

  // Destructor
  ~BASIS_REF_CONST(){}
};

// Basis functions on the reference element (used on parallel prisms)
class BASIS_REF_LIN : public BASIS_REF
{
protected:
  // Compute values of basis functions at a point
  // Parameters: xi - point coordinates; result - values of basis functions
  void basis_func(const real xi[], func_type result) const
  {
    static long_real sqrt2=std::sqrt(2.), sqrt3=std::sqrt(3.), sqrt6=std::sqrt(6.);

    result[3]=sqrt6*(2.*xi[2]-1.);
    result[2]=2.*sqrt3*(1.-xi[0]-2.*xi[1]);
    result[1]=2.-6.*xi[0];
    result[0]=sqrt2;
  }

  // Compute values of gradients of basis functions at a point
  // Parameters: xi - point coordinates; result - values of gradients of basis functions
  void basis_grad(const real xi[], grad_type result) const
  {
    static long_real sqrt3=std::sqrt(3.), sqrt6=std::sqrt(6.);

    result[3][0]=result[3][1]=0.; result[3][2]=2.*sqrt6;
    result[2][0]=-2.*sqrt3; result[2][1]=-4.*sqrt3; result[2][2]=0.;
    result[1][0]=-6.; result[1][1]=result[1][2]=0.;
    result[0][0]=result[0][1]=result[0][2]=0.;
  }

  // Compute values of integrals in the z-direction of basis functions at a point
  // Parameters: xi - point coordinates;
  //             result - values of integrals in the z-direction of basis functions
  void basis_func_vert_integral(const real xi[], func_type result) const
  {
    static long_real sqrt2=std::sqrt(2.), sqrt3=std::sqrt(3.), sqrt6=std::sqrt(6.);

    result[3]=sqrt6*(xi[2]-1.)*xi[2];
    result[2]=2.*sqrt3*(1.-xi[0]-2.*xi[1])*xi[2];
    result[1]=(2.-6.*xi[0])*xi[2];
    result[0]=sqrt2*xi[2];
  }

public:
  // Default constructor
  BASIS_REF_LIN() {Dim()=4;}

  // Destructor
  ~BASIS_REF_LIN(){}
};

// Basis functions on the reference element (used on parallel prisms)
class BASIS_REF_QUAD : public BASIS_REF
{
protected:
  // Compute values of basis functions at a point
  // Parameters: xi - point coordinates; result - values of basis functions
  void basis_func(const real xi[], func_type result) const
  {
    static long_real sqrt2=std::sqrt(2.), sqrt3=std::sqrt(3.), 
      three_sqrt5=3.*std::sqrt(5.), sqrt6=std::sqrt(6.), sqrt10=std::sqrt(10.);

    result[9]=6.*(1.-xi[0]-2.*xi[1])*(2.*xi[2]-1.);
    result[8]=sqrt3*(2.-6.*xi[0])*(2.*xi[2]-1.);
    result[7]=sqrt10*(6.*xi[2]*(xi[2]-1.)+1.);
    result[6]=three_sqrt5*((3.*xi[0]+8.*xi[1]-4.)*xi[0]+(3.*xi[1]-4.)*xi[1]+1.);
    result[5]=sqrt3*((5.*xi[0]-4.)*xi[0]+(-15.*xi[1]+12.)*xi[1]-1.);
    result[4]=sqrt6*((10.*xi[0]-8.)*xi[0]+1.);
    result[3]=sqrt6*(2.*xi[2]-1.);
    result[2]=2.*sqrt3*(1.-xi[0]-2.*xi[1]);
    result[1]=2.-6.*xi[0];
    result[0]=sqrt2;
  }

  // Compute values of gradients of basis functions at a point
  // Parameters: xi - point coordinates; result - values of gradients of basis functions
  void basis_grad(const real xi[], grad_type result) const
  {
    static long_real sqrt2=std::sqrt(2.), sqrt3=std::sqrt(3.), 
      three_sqrt5=3.*std::sqrt(5.), sqrt6=std::sqrt(6.), sqrt10=std::sqrt(10.);

    result[9][0]=-6.*(2.*xi[2]-1.); result[9][1]=-12.*(2.*xi[2]-1.); 
    result[9][2]=12.*(1.-xi[0]-2.*xi[1]);
    result[8][0]=-6.*sqrt3*(2.*xi[2]-1.); result[8][1]=0.; result[8][2]=2.*sqrt3*(2.-6.*xi[0]);
    result[7][0]=result[7][1]=0.; result[7][2]=sqrt10*(12.*xi[2]-6.);
    result[6][0]=three_sqrt5*(6.*xi[0]+8.*xi[1]-4.); 
    result[6][1]=three_sqrt5*(8.*xi[0]+6.*xi[1]-4.); result[6][2]=0.;
    result[5][0]=sqrt3*(10.*xi[0]-4.); result[5][1]=sqrt3*(-30.*xi[1]+12.); result[5][2]=0.;
    result[4][0]=sqrt6*(20.*xi[0]-8.); result[4][1]=result[4][2]=0.;
    result[3][0]=result[3][1]=0.; result[3][2]=2.*sqrt6;
    result[2][0]=-2.*sqrt3; result[2][1]=-4.*sqrt3; result[2][2]=0.;
    result[1][0]=-6.; result[1][1]=result[1][2]=0.;
    result[0][0]=result[0][1]=result[0][2]=0.;
  }

  // Compute values of integrals in the z-direction of basis functions at a point
  // Parameters: xi - point coordinates;
  //             result - values of integrals in the z-direction of basis functions
  void basis_func_vert_integral(const real xi[], func_type result) const
  {
    static long_real sqrt2=std::sqrt(2.), sqrt3=std::sqrt(3.), 
      three_sqrt5=3.*std::sqrt(5.), sqrt6=std::sqrt(6.), sqrt10=std::sqrt(10.);

    result[9]=6.*(1.-xi[0]-2.*xi[1])*(xi[2]-1.)*xi[2];
    result[8]=sqrt3*(2.-6.*xi[0])*(xi[2]-1.)*xi[2];
    result[7]=sqrt10*(xi[2]*(2.*xi[2]-3.)+1.)*xi[2];
    result[6]=three_sqrt5*((3.*xi[0]+8.*xi[1]-4.)*xi[0]+(3.*xi[1]-4.)*xi[1]+1.)*xi[2];
    result[5]=sqrt3*((5.*xi[0]-4.)*xi[0]+(-15.*xi[1]+12.)*xi[1]-1.)*xi[2];
    result[4]=sqrt6*((10.*xi[0]-8.)*xi[0]+1.)*xi[2];
    result[3]=sqrt6*(xi[2]-1.)*xi[2];
    result[2]=2.*sqrt3*(1.-xi[0]-2.*xi[1])*xi[2];
    result[1]=(2.-6.*xi[0])*xi[2];
    result[0]=sqrt2*xi[2];
  }

public:
  // Default constructor
  BASIS_REF_QUAD() {Dim()=10;}

  // Destructor
  ~BASIS_REF_QUAD(){}
};



//--------------------------------------------------------------------------------------
// BASIS_PHYS Class: Defines basis functions in pshysical space for 3D elements
//                   used in nonparallel prisms
//--------------------------------------------------------------------------------------

class BASIS_PHYS : public BASIS
{
protected:
  real ref_point[d_space], scale[d_space]; // Scaling parameters

public:
  // Default constructor
  BASIS_PHYS() {ref_point[0]=ref_point[1]=ref_point[2]=scale[0]=scale[1]=scale[2]=0.;}

  // Constructor
  // Parameters: mapping - map object from the reference prism
  BASIS_PHYS(const MapPrism& mapping) {mapping.GetPhysicalBasisParam(scale, ref_point);}

  // Constructor
  // Parameters: in - hot input file stream
  BASIS_PHYS(std::ifstream& in)
  {in.read((char*)ref_point, sizeof ref_point); in.read((char*)scale, sizeof scale);}

  // Destructor
  ~BASIS_PHYS() {}

  // Compute values of basis functions at a point in physical space
  // Parameters: x - point coordinates; mapping - map object from the reference element;
  //             result - values of basis functions
  void basis_func_phys(const real x[], const MAP& mapping, func_type result) const
  {basis_func(x, result);}

  // Compute values of gradients of basis functions at a point in physical space
  // Parameters: x - point coordinates; mapping - map object from the reference element;
  //             result - values of gradients of basis functions
  void basis_grad_phys(const real x[], const MAP& mapping, grad_type result) const
  {basis_grad(x, result);}

  // Compute values of basis functions at a point in reference space
  // Parameters: xi - point coordinates; mapping - map object from the reference element;
  //             result - values of basis functions
  void basis_func_ref(const real xi[], const MAP& mapping, func_type result) const
  {real x[d_space]; mapping.mapto(xi, x); basis_func(x, result);}

  // Compute values of gradients of basis functions at a point in reference space
  // Parameters: xi - point coordinates; mapping - map object from the reference element;
  //             result - values of gradients of basis functions
  void basis_grad_ref(const real xi[], const MAP& mapping, grad_type result) const
  {real x[d_space]; mapping.mapto(xi, x); basis_grad(x, result);}

  // Compute values of basis functions at a point in reference space
  // Parameters: xi - point coordinates; result - values of basis functions
  void basis_func_ref(const real xi[], func_type result) const
  {std::cerr << "File basis.H: Calling a physical basis function without mapping!\n"; std::exit(1);}

  // Compute values of gradients of basis functions at a point in reference space
  // Parameters: xi - point coordinates; result - values of gradients of basis functions
  void basis_grad_ref(const real xi[], grad_type result) const
  {std::cerr << "File basis.H: Calling a physical basis function without mapping!\n"; std::exit(1);}

  // Compute values of integrals in the z-direction of basis functions at a point
  // Parameters: x - point coordinates; mapping - map object from the reference element;
  //             result - values of integrals in the z-direction of basis functions
  void basis_func_vert_integral_phys(const real x[], const MAP& mapping, func_type result) const
  {
    real xi[d_space], x_top[d_space]={x[0], x[1]};
    func_type result_p;

    mapping.mapto_inverse(x, xi);
    xi[2]=0.;
    x_top[2]=((MapPrism&)mapping).mapto_z(xi);
    basis_func_vert_integral(x, result_p);
    basis_func_vert_integral(x_top, result);

    for (int j=0; j<Dim(); j++)
      result[j]-=result_p[j];
  }
//   void basis_func_vert_integral_phys(const real x[], const MAP& mapping, func_type result) const
//   {
//     real xi[d_space], x_bottom[d_space]={x[0], x[1]};
//     func_type result_bottom;

//     mapping.mapto_inverse(x, xi);
//     xi[2]=1.;
//     x_bottom[2]=((MapPrism&)mapping).mapto_z(xi);
//     basis_func_vert_integral(x, result);
//     basis_func_vert_integral(x_bottom, result_bottom);

//     for (int j=0; j<Dim(); j++)
//       result[j]-=result_bottom[j];
//   }

  // Hot file output for scaling parameters used in the physical basis
  friend std::ofstream& operator<<(std::ofstream& out, const BASIS_PHYS& basis)
  {out.write((char*)basis.ref_point, sizeof basis.ref_point);
  out.write((char*)basis.scale, sizeof basis.scale); return out;}
};

// Basis functions on the physical element (used on non-parallel prisms)
class BASIS_PHYS_CONST : public BASIS_PHYS
{
protected:
  // Compute values of basis functions at a point
  // Parameters: xi - point coordinates; result - values of basis functions
  void basis_func(const real xi[d_space], func_type result) const
  {
    result[0]=1.;
  }

  // Compute values of gradients of basis functions at a point
  // Parameters: xi - point coordinates; result - values of gradients of basis functions
  void basis_grad(const real xi[d_space], grad_type result) const
  {
    result[0][0]=result[0][1]=result[0][2]=0.;
  }

  // Compute values of integrals in the z-direction of basis functions at a point
  // Parameters: xi - point coordinates;
  //             result - values of integrals in the z-direction of basis functions
  void basis_func_vert_integral(const real xi[], func_type result) const
  {
    result[0]=xi[2];
  }

public:
  // Default constructor
  BASIS_PHYS_CONST() : BASIS_PHYS() {Dim()=1;}

  // Constructor
  // Parameters: mapping - map object from the reference prism
  BASIS_PHYS_CONST(const MapPrism& mapping) : BASIS_PHYS(mapping) {Dim()=1;}

  // Constructor
  // Parameters: in - hot input file stream
  BASIS_PHYS_CONST(std::ifstream& in) : BASIS_PHYS(in) {Dim()=1;}

  // Destructor
  ~BASIS_PHYS_CONST() {}
};

// Basis functions on the physical element (used on non-parallel prisms)
class BASIS_PHYS_LIN : public BASIS_PHYS
{
protected:
  // Compute values of basis functions at a point
  // Parameters: xi - point coordinates; result - values of basis functions
  void basis_func(const real xi[d_space], func_type result) const
  {
    real x[d_space];

    for (int i=0; i<d_space; x[i]=(xi[i]-ref_point[i])/scale[i], i++);
    result[3]=x[2];
    result[2]=x[1];
    result[1]=x[0];
    result[0]=1.;
  }

  // Compute values of gradients of basis functions at a point
  // Parameters: xi - point coordinates; result - values of gradients of basis functions
  void basis_grad(const real xi[d_space], grad_type result) const
  {
    result[3][2]=1./scale[2], result[3][0]=result[3][1]=0.;
    result[2][1]=1./scale[1], result[2][0]=result[2][2]=0.;
    result[1][0]=1./scale[0], result[1][1]=result[1][2]=0.;
    result[0][0]=result[0][1]=result[0][2]=0.;
  }

  // Compute values of integrals in the z-direction of basis functions at a point
  // Parameters: xi - point coordinates;
  //             result - values of integrals in the z-direction of basis functions
  void basis_func_vert_integral(const real xi[], func_type result) const
  {
    real x[d_space];

    for (int i=0; i<d_space; x[i]=(xi[i]-ref_point[i])/scale[i], i++);
    result[3]=0.5*x[2]*xi[2];
    result[2]=x[1]*xi[2];
    result[1]=x[0]*xi[2];
    result[0]=xi[2];
  }

public:
  // Default constructor
  BASIS_PHYS_LIN() : BASIS_PHYS() {Dim()=4;}

  // Constructor
  // Parameters: mapping - map object from the reference prism
  BASIS_PHYS_LIN(const MapPrism& mapping) : BASIS_PHYS(mapping) {Dim()=4;}

  // Constructor
  // Parameters: in - hot input file stream
  BASIS_PHYS_LIN(std::ifstream& in) : BASIS_PHYS(in) {Dim()=4;}

  // Destructor
  ~BASIS_PHYS_LIN() {}
};

// Basis functions on the physical element (used on non-parallel prisms)
class BASIS_PHYS_QUAD : public BASIS_PHYS
{
protected:
  // Compute values of basis functions at a point
  // Parameters: xi - point coordinates; result - values of basis functions
  void basis_func(const real xi[d_space], func_type result) const
  {
    real x[d_space];

    for (int i=0; i<d_space; x[i]=(xi[i]-ref_point[i])/scale[i], i++);
    result[9]=x[2]*x[1];
    result[8]=x[2]*x[0];
    result[7]=x[0]*x[1];
    result[6]=x[2]*x[2];
    result[5]=x[1]*x[1];
    result[4]=x[0]*x[0];
    result[3]=x[2];
    result[2]=x[1];
    result[1]=x[0];
    result[0]=1.;
  }

  // Compute values of gradients of basis functions at a point
  // Parameters: xi - point coordinates; result - values of gradients of basis functions
  void basis_grad(const real xi[d_space], grad_type result) const
  {
    real x[d_space];

    for (int i=0; i<d_space; x[i]=(xi[i]-ref_point[i])/scale[i], i++);
    result[9][0]=0., result[9][1]=x[2]/scale[1], result[9][2]=x[1]/scale[2];
    result[8][0]=x[2]/scale[0], result[8][1]=0., result[8][2]=x[0]/scale[2];
    result[7][0]=x[1]/scale[0], result[7][1]=x[0]/scale[1], result[7][2]=0.;
    result[6][0]=result[6][1]=0., result[6][2]=2.*x[2]/scale[2];
    result[5][1]=2.*x[1]/scale[1], result[5][0]=result[5][2]=0.;
    result[4][0]=2.*x[0]/scale[0], result[4][1]=result[4][2]=0.;
    result[3][2]=1./scale[2], result[3][0]=result[3][1]=0.;
    result[2][1]=1./scale[1], result[2][0]=result[2][2]=0.;
    result[1][0]=1./scale[0], result[1][1]=result[1][2]=0.;
    result[0][0]=result[0][1]=result[0][2]=0.;
  }

  // Compute values of integrals in the z-direction of basis functions at a point
  // Parameters: xi - point coordinates;
  //             result - values of integrals in the z-direction of basis functions
  void basis_func_vert_integral(const real xi[], func_type result) const
  {
    real x[d_space];

    for (int i=0; i<d_space; x[i]=(xi[i]-ref_point[i])/scale[i], i++);
    result[9]=0.5*x[2]*x[1]*xi[2];
    result[8]=0.5*x[2]*x[0]*xi[2];
    result[7]=x[0]*x[1]*xi[2];
    result[6]=1./3.*x[2]*x[2]*xi[2];
    result[5]=x[1]*x[1]*xi[2];
    result[4]=x[0]*x[0]*xi[2];
    result[3]=0.5*x[2]*xi[2];
    result[2]=x[1]*xi[2];
    result[1]=x[0]*xi[2];
    result[0]=xi[2];
  }

public:
  // Default constructor
  BASIS_PHYS_QUAD() : BASIS_PHYS() {Dim()=10;}

  // Constructor
  // Parameters: mapping - map object from the reference prism
  BASIS_PHYS_QUAD(const MapPrism& mapping) : BASIS_PHYS(mapping) {Dim()=10;}

  // Constructor
  // Parameters: in - hot input file stream
  BASIS_PHYS_QUAD(std::ifstream& in) : BASIS_PHYS(in) {Dim()=10;}

  // Destructor
  ~BASIS_PHYS_QUAD() {}
};

//--------------------------------------------------------------------------------------
// BASIS_ZPHYS Class: Defines basis functions in pshysical space for 3D elements
//                   used in nonparallel prisms
//--------------------------------------------------------------------------------------

class BASIS_ZPHYS : public BASIS
{
protected:
  real ref_point, scale; // Scaling parameters

public:
  // Default constructor
  BASIS_ZPHYS() {ref_point=scale=0.;}

  // Constructor
  // Parameters: mapping - map object from the reference prism
  BASIS_ZPHYS(const MapPrism& mapping)
  {
    real xi[d_space]={1./3., 1./3., 0.5}, x[d_space];

    mapping.mapto(xi, x);
    ref_point=x[2];
    scale=0.5*mapping.GetVerticalScale();
  }

  // Constructor
  // Parameters: in - hot input file stream
  BASIS_ZPHYS(std::ifstream& in)
  {in.read((char*)&ref_point, sizeof ref_point); in.read((char*)&scale, sizeof scale);}

  // Destructor
  ~BASIS_ZPHYS() {}

  // Compute values of basis functions at a point in physical space
  // Parameters: xi - point coordinates; mapping - map object from the reference element;
  //             result - values of basis functions
  void basis_func_phys(const real x[], const MAP& mapping, func_type result) const
  {real xi[d_space]; ((MapPrism&)mapping).Map2D().mapto_inverse(x, xi);
    xi[2]=(x[2]-ref_point)/scale; basis_func(xi, result);}

  // Compute values of gradients of basis functions at a point in physical space
  // Parameters: x - point coordinates; mapping - map object from the reference element;
  //             result - values of gradients of basis functions
  void basis_grad_phys(const real x[], const MAP& mapping, grad_type result) const
  {
    real xi[d_space];
    grad_type grad_tmp;

    ((MapPrism&)mapping).Map2D().mapto_inverse(x, xi);
    xi[2]=(x[2]-ref_point)/scale;
    basis_grad(xi, grad_tmp);
    for (int j=0; j<Dim(); j++)
    {
      ((MapPrism&)mapping).Map2D().MapGradient(grad_tmp[j], result[j]);
      result[j][2]=grad_tmp[j][2];
    }
  }

  // Compute values of basis functions at a point in reference space
  // Parameters: xi - point coordinates; mapping - map object from the reference element;
  //             result - values of basis functions
  void basis_func_ref(const real xi[], const MAP& mapping, func_type result) const
  {real x[d_space]={xi[0], xi[1]}; 
    x[2]=(((MapPrism&)mapping).mapto_z(xi)-ref_point)/scale; basis_func(x, result);}

  // Compute values of gradients of basis functions at a point in reference space
  // Parameters: xi - point coordinates; mapping - map object from the reference element;
  //             result - values of gradients of basis functions
  void basis_grad_ref(const real xi[], const MAP& mapping, grad_type result) const
  {
    real x[d_space]={xi[0], xi[1]};
    grad_type grad_tmp;

    x[2]=(((MapPrism&)mapping).mapto_z(xi)-ref_point)/scale;
    basis_grad(x, grad_tmp);
    for (int j=0; j<Dim(); j++)
    {
      ((MapPrism&)mapping).Map2D().MapGradient(grad_tmp[j], result[j]);
      result[j][2]=grad_tmp[j][2];
    }
  }

  // Compute values of basis functions at a point in reference space
  // Parameters: xi - point coordinates; result - values of basis functions
  void basis_func_ref(const real xi[], func_type result) const
  {std::cerr << "File basis.H: Calling a physical basis function without mapping!\n"; std::exit(1);}

  // Compute values of gradients of basis functions at a point in reference space
  // Parameters: xi - point coordinates; result - values of gradients of basis functions
  void basis_grad_ref(const real xi[], grad_type result) const
  {std::cerr << "File basis.H: Calling a physical basis function without mapping!\n"; std::exit(1);}

  // Compute values of integrals in the z-direction of basis functions at a point
  // Parameters: x - point coordinates; mapping - map object from the reference element;
  //             result - values of integrals in the z-direction of basis functions
  void basis_func_vert_integral_phys(const real x[], const MAP& mapping, func_type result) const
  {
    real xi[d_space];
    func_type result_p;

    ((MapPrism&)mapping).Map2D().mapto_inverse(x, xi);
    real xi_top[d_space]={xi[0], xi[1]};
    xi[2]=0.;
    xi_top[2]=(((MapPrism&)mapping).mapto_z(xi)-ref_point)/scale;
    xi[2]=(x[2]-ref_point)/scale;
    basis_func_vert_integral(xi, result_p);
    basis_func_vert_integral(xi_top, result);

    for (int j=0; j<Dim(); j++)
      result[j]-=result_p[j];
  }
//   void basis_func_vert_integral_phys(const real x[], const MAP& mapping, func_type result) const
//   {
//     real xi[d_space];
//     func_type result_bottom;

//     ((MapPrism&)mapping).Map2D().mapto_inverse(x, xi);
//     xi[2]=1.;
//     real xi_bottom[d_space]={xi[0], xi[1]};
//     xi_bottom[2]=(((MapPrism&)mapping).mapto_z(xi)-ref_point)/scale;
//     xi[2]=(x[2]-ref_point)/scale;
//     basis_func_vert_integral(xi, result);
//     basis_func_vert_integral(xi_bottom, result_bottom);

//     for (int j=0; j<Dim(); j++)
//       result[j]-=result_bottom[j];
//   }

  // Hot file output for scaling parameters used in the physical basis
  friend std::ofstream& operator<<(std::ofstream& out, const BASIS_ZPHYS& basis)
  {out.write((char*)&basis.ref_point, sizeof basis.ref_point);
  out.write((char*)&basis.scale, sizeof basis.scale); return out;}
};

// Basis functions on the physical element (used on non-parallel prisms)
class BASIS_ZPHYS_CONST : public BASIS_ZPHYS
{
protected:
  // Compute values of basis functions at a point
  // Parameters: xi - point coordinates; result - values of basis functions
  void basis_func(const real xi[d_space], func_type result) const
  {
    static long_real sqrt2=std::sqrt(2.);

    result[0]=sqrt2;
  }

  // Compute values of gradients of basis functions at a point
  // Parameters: xi - point coordinates; result - values of gradients of basis functions
  void basis_grad(const real xi[d_space], grad_type result) const
  {
    result[0][0]=result[0][1]=result[0][2]=0.;
  }

  // Compute values of integrals in the z-direction of basis functions at a point
  // Parameters: x - point coordinates;
  //             result - values of integrals in the z-direction of basis functions
  void basis_func_vert_integral(const real xi[], func_type result) const
  {
    static long_real sqrt2=std::sqrt(2.);

    result[0]=sqrt2*xi[2]*scale;
  }

public:
  // Default constructor
  BASIS_ZPHYS_CONST() : BASIS_ZPHYS() {Dim()=1;}

  // Constructor
  // Parameters: mapping - map object from the reference prism
  BASIS_ZPHYS_CONST(const MapPrism& mapping) : BASIS_ZPHYS(mapping) {Dim()=1;}

  // Constructor
  // Parameters: in - hot input file stream
  BASIS_ZPHYS_CONST(std::ifstream& in) : BASIS_ZPHYS(in) {Dim()=1;}

  // Destructor
  ~BASIS_ZPHYS_CONST() {}
};

// Basis functions on the physical element (used on non-parallel prisms)
class BASIS_ZPHYS_LIN : public BASIS_ZPHYS
{
protected:
  // Compute values of basis functions at a point
  // Parameters: xi - point coordinates; result - values of basis functions
  void basis_func(const real xi[d_space], func_type result) const
  {
    static long_real sqrt2=std::sqrt(2.), sqrt3=std::sqrt(3.), sqrt6=std::sqrt(6.);

    result[3]=sqrt6*(2.*xi[2]-1.);
    result[2]=2.*sqrt3*(1.-xi[0]-2.*xi[1]);
    result[1]=2.-6.*xi[0];
    result[0]=sqrt2;
  }

  // Compute values of gradients of basis functions at a point
  // Parameters: xi - point coordinates; result - values of gradients of basis functions
  void basis_grad(const real xi[d_space], grad_type result) const
  {
    static long_real sqrt3=std::sqrt(3.), sqrt6=std::sqrt(6.);

    result[3][0]=result[3][1]=0.; result[3][2]=2.*sqrt6/scale;
    result[2][0]=-2.*sqrt3; result[2][1]=-4.*sqrt3; result[2][2]=0.;
    result[1][0]=-6.; result[1][1]=result[1][2]=0.;
    result[0][0]=result[0][1]=result[0][2]=0.;
  }

  // Compute values of integrals in the z-direction of basis functions at a point
  // Parameters: xi - point coordinates;
  //             result - values of integrals in the z-direction of basis functions
  void basis_func_vert_integral(const real xi[], func_type result) const
  {
    static long_real sqrt2=std::sqrt(2.), sqrt3=std::sqrt(3.), sqrt6=std::sqrt(6.);

    result[3]=sqrt6*(xi[2]-1.)*xi[2]*scale;
    result[2]=2.*sqrt3*(1.-xi[0]-2.*xi[1])*xi[2]*scale;
    result[1]=(2.-6.*xi[0])*xi[2]*scale;
    result[0]=sqrt2*xi[2]*scale;
  }

public:
  // Default constructor
  BASIS_ZPHYS_LIN() : BASIS_ZPHYS() {Dim()=4;}

  // Constructor
  // Parameters: mapping - map object from the reference prism
  BASIS_ZPHYS_LIN(const MapPrism& mapping) : BASIS_ZPHYS(mapping) {Dim()=4;}

  // Constructor
  // Parameters: in - hot input file stream
  BASIS_ZPHYS_LIN(std::ifstream& in) : BASIS_ZPHYS(in) {Dim()=4;}

  // Destructor
  ~BASIS_ZPHYS_LIN() {}
};

// Basis functions on the physical element (used on non-parallel prisms)
class BASIS_ZPHYS_QUAD : public BASIS_ZPHYS
{
protected:
  // Compute values of basis functions at a point
  // Parameters: xi - point coordinates; result - values of basis functions
  void basis_func(const real xi[d_space], func_type result) const
  {
    static long_real sqrt2=std::sqrt(2.), sqrt3=std::sqrt(3.), 
      three_sqrt5=3.*std::sqrt(5.), sqrt6=std::sqrt(6.), sqrt10=std::sqrt(10.);

    result[9]=6.*(1.-xi[0]-2.*xi[1])*(2.*xi[2]-1.);
    result[8]=sqrt3*(2.-6.*xi[0])*(2.*xi[2]-1.);
    result[7]=sqrt10*(6.*xi[2]*(xi[2]-1.)+1.);
    result[6]=three_sqrt5*((3.*xi[0]+8.*xi[1]-4.)*xi[0]+(3.*xi[1]-4.)*xi[1]+1.);
    result[5]=sqrt3*((5.*xi[0]-4.)*xi[0]+(-15.*xi[1]+12.)*xi[1]-1.);
    result[4]=sqrt6*((10.*xi[0]-8.)*xi[0]+1.);
    result[3]=sqrt6*(2.*xi[2]-1.);
    result[2]=2.*sqrt3*(1.-xi[0]-2.*xi[1]);
    result[1]=2.-6.*xi[0];
    result[0]=sqrt2;
  }

  // Compute values of gradients of basis functions at a point
  // Parameters: xi - point coordinates; result - values of gradients of basis functions
  void basis_grad(const real xi[d_space], grad_type result) const
  {
    static long_real sqrt2=std::sqrt(2.), sqrt3=std::sqrt(3.), 
      three_sqrt5=3.*std::sqrt(5.), sqrt6=std::sqrt(6.), sqrt10=std::sqrt(10.);

    result[9][0]=-6.*(2.*xi[2]-1.); result[9][1]=-12.*(2.*xi[2]-1.); 
    result[9][2]=12.*(1.-xi[0]-2.*xi[1])/scale;
    result[8][0]=-6.*sqrt3*(2.*xi[2]-1.); result[8][1]=0.; result[8][2]=2.*sqrt3*(2.-6.*xi[0])/scale;
    result[7][0]=result[7][1]=0.; result[7][2]=sqrt10*(12.*xi[2]-6.)/scale;
    result[6][0]=three_sqrt5*(6.*xi[0]+8.*xi[1]-4.); 
    result[6][1]=three_sqrt5*(8.*xi[0]+6.*xi[1]-4.); result[6][2]=0.;
    result[5][0]=sqrt3*(10.*xi[0]-4.); result[5][1]=sqrt3*(-30.*xi[1]+12.); result[5][2]=0.;
    result[4][0]=sqrt6*(20.*xi[0]-8.); result[4][1]=result[4][2]=0.;
    result[3][0]=result[3][1]=0.; result[3][2]=2.*sqrt6/scale;
    result[2][0]=-2.*sqrt3; result[2][1]=-4.*sqrt3; result[2][2]=0.;
    result[1][0]=-6.; result[1][1]=result[1][2]=0.;
    result[0][0]=result[0][1]=result[0][2]=0.;
  }

  // Compute values of integrals in the z-direction of basis functions at a point
  // Parameters: xi - point coordinates;
  //             result - values of integrals in the z-direction of basis functions
  void basis_func_vert_integral(const real xi[], func_type result) const
  {
    static long_real sqrt2=std::sqrt(2.), sqrt3=std::sqrt(3.), 
      three_sqrt5=3.*std::sqrt(5.), sqrt6=std::sqrt(6.), sqrt10=std::sqrt(10.);

    result[9]=6.*(1.-xi[0]-2.*xi[1])*(xi[2]-1.)*xi[2]*scale;
    result[8]=sqrt3*(2.-6.*xi[0])*(xi[2]-1.)*xi[2]*scale;
    result[7]=sqrt10*(xi[2]*(2.*xi[2]-3.)+1.)*xi[2]*scale;
    result[6]=three_sqrt5*((3.*xi[0]+8.*xi[1]-4.)*xi[0]+(3.*xi[1]-4.)*xi[1]+1.)*xi[2]*scale;
    result[5]=sqrt3*((5.*xi[0]-4.)*xi[0]+(-15.*xi[1]+12.)*xi[1]-1.)*xi[2]*scale;
    result[4]=sqrt6*((10.*xi[0]-8.)*xi[0]+1.)*xi[2]*scale;
    result[3]=sqrt6*(xi[2]-1.)*xi[2]*scale;
    result[2]=2.*sqrt3*(1.-xi[0]-2.*xi[1])*xi[2]*scale;
    result[1]=(2.-6.*xi[0])*xi[2]*scale;
    result[0]=sqrt2*xi[2]*scale;
  }

public:
  // Default constructor
  BASIS_ZPHYS_QUAD() : BASIS_ZPHYS() {Dim()=10;}

  // Constructor
  // Parameters: mapping - map object from the reference prism
  BASIS_ZPHYS_QUAD(const MapPrism& mapping) : BASIS_ZPHYS(mapping) {Dim()=10;}

  // Constructor
  // Parameters: in - hot input file stream
  BASIS_ZPHYS_QUAD(std::ifstream& in) : BASIS_ZPHYS(in) {Dim()=10;}

  // Destructor
  ~BASIS_ZPHYS_QUAD() {}
};
#endif
