//----------------------------------------------------------------------------------------
// mesh_3d_sw.cpp: Contains definitions for 3D node, face, and prism shallow-water objects
//----------------------------------------------------------------------------------------

#include <iostream>
#include <fstream>
#include <cassert>
#include <cstring>
#include <deque>
#include "general.H"
#include "les_solver.H"
#include "mesh_3d_sw.H"
#include "mesh_container.H"
#include "pde.H"
#ifdef TEST_KEY
#include "test.H"
#endif

extern PROBLEM_DESCRIPTION problem_info;          // Storage for problem parameters
extern QUAD_CONTAINER quadrature_index;           // Quadrature objects
extern BASIS_2D *basis_2d[max_order+1];           // 2D basis objects
extern BASIS_REF *basis_ref[max_order+1];         // 3D basis objects for parallel prisms
extern std::deque<BASIS_ZPHYS_CONST> basis_phys_const;  // 3D basis objects for nonparallel prisms
extern std::deque<BASIS_ZPHYS_LIN> basis_phys_lin;      // 3D basis objects for nonparallel prisms
extern std::deque<BASIS_ZPHYS_QUAD> basis_phys_quad;    // 3D basis objects for nonparallel prisms
extern MESH mesh;                                // Mesh object



//--------------------------------------------------------------------------------------
//  FACE_HORIZ Class: Horizontal face class for shallow-water problems
//--------------------------------------------------------------------------------------

// Default constructor
// Parameters: elem_2d - corresponding 2D element
FACE_HORIZ_SW::FACE_HORIZ_SW(TRIANGLE_2D_SW& elem_2d) : FACE_HORIZ(elem_2d)
{
  container_2d=&elem_2d;
  coef_scale_d[0]=coef_scale_d[1]=0.;
  n_basis[0]=n_basis[1]=0;
  for (int i=0; i<d_var; basis_index[0][i]=basis_index[1][i]=BAD_INDEX, i++);
  // Bottom of the reference element maps to the top of the physical and vice versa
  quad=quadrature_index.GetHorizFaceQuad(problem_info.def_order);
}

// Constructor for hot start input
// Parameters: in - hot start input file stream
FACE_HORIZ_SW::FACE_HORIZ_SW(std::ifstream& in) : FACE_HORIZ(in)
{
  in.read((char*)coef_scale_d, sizeof coef_scale_d);
  in.read((char*)n_basis, sizeof n_basis);
  in.read((char*)basis_index, sizeof basis_index);
  container_2d=&mesh.triangle_2d_list[Element2D()];
  // Bottom of the reference element maps to the top of the physical and vice versa
  quad=quadrature_index.GetHorizFaceQuad(problem_info.def_order);
}

// Compute mapping from the reference face
void FACE_HORIZ_SW::ComputeReferenceMapping()
{
  const real* coord[vert_h];

  for (int i=0; i<vert_h; coord[i]=mesh.node_3d_list[Node(i)].Coord(), i++);
  FACE_HORIZ::ComputeReferenceMapping(coord);
}

// Return value: pointer to matrix for "W" compuutation, lvalue
long_real* FACE_HORIZ_SW::w_matrix() const
{
  assert_debug(IsElement(1));
  return mesh.prism_list[Element(1)].w_matrix();
}

// Update recursively face-element connectivity info
// Parameters: el_ind - element index in the face; el_new - element id
void FACE_HORIZ_SW::SetElementRecursive(int el_ind, int el_new)
{
  Element(el_ind)=el_new;
  for (int i=0; i<NChildren(); i++)
    mesh.face_horiz_list[Child(i)].SetElementRecursive(el_ind, el_new);
}

// Set recursively the face (or its descendants) to be active
void FACE_HORIZ_SW::SetActiveRecursive()
{
  bool active = IsInterior() ? IsElement(0) && IsElement(1) : true;

  for (int k=0; k<2; k++)
    active = IsElement(k) ? active && mesh.prism_list[Element(k)].IsActive() : active;
  if (active)
  {
    Status()|=ACTIVE;
    UnSetUpdated();
    for (int i=0; i<NChildren(); i++)
      mesh.face_horiz_list[Child(i)].UnSetActiveRecursive();
  }
  else
  {
    Status()&=~ACTIVE;
    for (int i=0; i<NChildren(); i++)
      mesh.face_horiz_list[Child(i)].SetActiveRecursive();
  }
}

// // Set recursively the face (or its descendants) to be active
// // This version works for bottom faces and must be overloaded for interior and surface faces
// void FACE_HORIZ_SW::SetActiveRecursive()
// {
//   assert_debug(IsElement(0));
//   if (mesh.prism_list[Element(0)].IsActive())
//   {
//     Status()|=ACTIVE;
//     UnSetUpdated();
//   }
//   else
//   {
//     Status()&=~ACTIVE;
//     for (int i=0; i<NChildren(); i++)
//       mesh.face_horiz_list[Child(i)].SetActiveRecursive();
//   }
// }

// Set recursively the face (or its descendants) to be inactive
void FACE_HORIZ_SW::UnSetActiveRecursive()
{
  Status()&=~ACTIVE;
  for (int i=0; i<NChildren(); i++)
    mesh.face_horiz_list[Child(i)].UnSetActiveRecursive();
}

// Perform geometry update for moving free surface
void FACE_HORIZ_SW::UpdateSurfaceElevation()
{
  // Update numerical data
  ComputeReferenceMapping();
  UnSetUpdated();
}

#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
// Refine the face in four by edge bisection
// Parameters: node_new_ind - IDs of the new nodes, lvalue
// Return value: id of the first child face
void FACE_HORIZ_SW::Refine(int node_new_ind[vert_h])
{
  if (!IsRefined())
  {
    int n_children=mesh.triangle_2d_list[Element2D()].NChildren(),
      face_child_2d=mesh.triangle_2d_list[Element2D()].Child(0);
    TRIANGLE_2D *element_2d_child[4];
    FACE_HORIZ *face_horiz_child[4];
    NODE_3D *node_new[vert_h], *node[vert_h]=
      {&mesh.node_3d_list[Node(0)], &mesh.node_3d_list[Node(1)], &mesh.node_3d_list[Node(2)]};

    // Create new nodes
    for (int i=0; i<vert_h; i++)
    {
      if (node_new_ind[i]==BAD_INDEX)
      {
        int node_2d_ind=mesh.triangle_2d_list[face_child_2d+n_children-1].Node(i);
        NODE_3D_SW node_loc(mesh.node_2d_list[node_2d_ind]);

        node_new_ind[i]=node_loc.ID()=mesh.node_3d_list.size();
        if(IsSurface())
          node_loc.SetSurface();
        mesh.node_3d_list.push_back(node_loc);
      }
      node_new[i]=&mesh.node_3d_list[node_new_ind[i]];
    }
    // Create new faces
    for (int i=0; i<n_children; i++)
    {
      element_2d_child[i]=&mesh.triangle_2d_list[face_child_2d+i];
      FACE_HORIZ_SW* face_horiz_child_loc=CreateFaceHoriz(mesh.triangle_2d_list[face_child_2d+i]);

      face_horiz_child_loc->ID()=mesh.face_horiz_list.size();
      mesh.face_horiz_list.push_back(*face_horiz_child_loc);
      face_horiz_child[i]=&mesh.face_horiz_list[face_horiz_child_loc->ID()];
      delete face_horiz_child_loc;
    }
    FACE_HORIZ::Refine(element_2d_child, face_horiz_child, node, node_new);
    for (int i=0; i<n_children; i++)
    {
      FACE_HORIZ_SW* face_child=(FACE_HORIZ_SW*)face_horiz_child[i];

      face_child->ComputeReferenceMapping();
      face_child->InitializeStorage();
    }
  }
  else
  {
    for (int i=0; i<vert_h; i++)
      node_new_ind[i]=mesh.face_horiz_list[Child(NChildren()-1)].Node(i);
  }
}
#endif

// Initialize storage for data stored at quadrature points
void FACE_HORIZ_SW::InitializeStorage()
{
  int n_quad=Quad()->GetNQuad();
  std::size_t ind=mass_storage_.size();

  BaseStorageFaceHoriz::Initialize(n_quad);
  for (int k=0; k<2; k++)
    if (IsElement(k))
    {
      PRISM_SW* prism=&mesh.prism_list[Element(k)];

      n_basis[k]=prism->NBasis();
      for (int i=0; i<d_var; i++)
        basis_index[k][i]=prism->Basis_index((VAR)i);
      for (int j=0; j<n_basis[k]; j++)
      {
        int n_dim=prism->Basis(j)->Dim();

        mass_storage_.insert(mass_storage_.end(), n_quad*n_dim, 0);
        for (SetIndex(0); InRange(); ind+=n_dim, (*this)++)
          quad_local().basis_func_ind[k][j]=ind;
      }
    }
}

// Update data shared by all horizontal faces
// Parameters: var_new - global storage for degrees of freedom;
//             var_flux_new - global storage for degrees of freedom for the fluxes;
//             var_rhs_new - global storage for degrees of freedom for the right hand side
void FACE_HORIZ_SW::Update_static(DOF var_new[d_var], DOF var_flux_new[d_var][d_space], 
                                  DOF var_rhs_new[d_var])
{
  for (int i=0; i<d_var; i++)
  {
    temp_storage.Var_dof[i]=var_new[i].GetDofIterator();
    temp_storage.L_Var_dof[i]=var_rhs_new[i].GetDofIterator();
    for (int j=0; j<d_space; j++)
      temp_storage.Q_dof[i][j]=var_flux_new[i][j].GetDofIterator();        
  }
}


// Update rarely changed data at quadrature points (basis functions, coordinates, etc)
// This version works for bottom faces, must be overloaded for interior and surface faces
void FACE_HORIZ_SW::Update_basis_quad()
{
  assert_debug(IsInitialized() && IsElement(0) && container_2d->IsUpdated());
  PRISM_SW* prism=&mesh.prism_list[Element(0)];

  coef_scale(0)=-Mapping().GetScale()/prism->Mapping().GetScale();
  for (SetIndex(0); InRange(); (*this)++)
  {
    Mapping().mapto(Quad()->p(GetIndex()), coord_quad());
    for (int j=0; j<n_basis[0]; j++)
      prism->Basis(j)->basis_func_phys(coord_quad(), prism->Mapping(), basis_func_assign_quad(0, j));
  }
}

// // Update often changed data at quadrature points (unknowns)
// // Parameters: t - time
// void FACE_HORIZ_SW::Update_variables_quad(real t)
// {
//   static const VAR* ind=VAR_3D_DT_INDEX;
//   int boundary_ind=-1;

//   assert_debug(IsUpdated());
//   for (int k=0; k<2; k++)
//     if (IsElement(k))      
//       for (SetIndex(0); InRange(); (*this)++)
//         for (int i=0; i<d_var_3d_dt; i++)
//           quad_local().Var[k][ind[i]]=Var_dof(k, ind[i])*basis_func_quad(k, ind[i]);
//     else
//       boundary_ind=k;
//   if (boundary_ind==-1)
//     return;
//   for (SetIndex(0); InRange(); (*this)++)
//     BC_Dirichlet_quad(t, 0, quad_local().Var[boundary_ind]);
// }


#ifdef BAROCLINIC_KEY
// Parameters: el_ind - element index on the face
// Return value: pointer to values of vertically integrated basis functions
//               at the quadrature point, rvalue
const real* FACE_HORIZ_SW::basis_func_integrated_quad() const
{
  static func_type basis_func_integrated;
  PRISM_SW* prism = IsSurface() ? &mesh.prism_list[Element(1)] : &mesh.prism_list[Element(0)];

  prism->Basis_dof(RHO)->
    basis_func_vert_integral_phys(coord_quad(), prism->Mapping(), basis_func_integrated);
  return basis_func_integrated;
}
#endif

// Hot start output
std::ofstream& operator<<(std::ofstream& out, const FACE_HORIZ_SW& face)
{
  out << (FACE_HORIZ)face;
  out.write((char*)face.coef_scale_d, sizeof face.coef_scale_d);
  out.write((char*)face.n_basis, sizeof face.n_basis);
  out.write((char*)face.basis_index, sizeof face.basis_index);
  return out;
}


//----------------------------------------------------------------------------
//                        FACE_HORIZ_SURFACE CLASS
//----------------------------------------------------------------------------

// Set recursively the face (or its descendants) to be active
// void FACE_HORIZ_SURFACE::SetActiveRecursive()
// {
//   assert_debug(IsElement(1));
//   if (mesh.prism_list[Element(1)].IsActive())
//   {
//     Status()|=ACTIVE;
//     UnSetUpdated();
//   }
//   else
//   {
//     Status()&=~ACTIVE;
//     for (int i=0; i<NChildren(); i++)
//       mesh.face_horiz_list[Child(i)].SetActiveRecursive();
//   }
// }

// Update rarely changed data at quadrature points (basis functions, coordinates, etc)
void FACE_HORIZ_SURFACE::Update_basis_quad()
{
  assert_debug(IsInitialized() && IsElement(1) && container_2d->IsUpdated());
  PRISM_SW* prism=&mesh.prism_list[Element(1)];

  coef_scale(1)=Mapping().GetScale()/prism->Mapping().GetScale();
  for (SetIndex(0); InRange(); (*this)++)
  {
    Mapping().mapto(Quad()->p(GetIndex()), coord_quad());
    container_2d->surface_elev_quad()=coord_quad(2);
    for (int j=0; j<n_basis[1]; j++)
      prism->Basis(j)->basis_func_phys(coord_quad(), prism->Mapping(), basis_func_assign_quad(1, j));
  }
}


//----------------------------------------------------------------------------
//                        FACE_HORIZ_SURFACE_STRESS CLASS
//----------------------------------------------------------------------------

// Set recursively the face (or its descendants) to be active
// void FACE_HORIZ_SURFACE_STRESS::SetActiveRecursive()
// {
//   assert_debug(IsElement(1));
//   if (mesh.prism_list[Element(1)].IsActive())
//   {
//     Status()|=ACTIVE;
//     UnSetUpdated();
//   }
//   else
//   {
//     Status()&=~ACTIVE;
//     for (int i=0; i<NChildren(); i++)
//       mesh.face_horiz_list[Child(i)].SetActiveRecursive();
//   }
// }

// Update rarely changed data at quadrature points (basis functions, coordinates, etc)
void FACE_HORIZ_SURFACE_STRESS::Update_basis_quad()
{
  assert_debug(IsInitialized() && IsElement(1) && container_2d->IsUpdated());
  PRISM_SW* prism=&mesh.prism_list[Element(1)];

  coef_scale(1)=Mapping().GetScale()/prism->Mapping().GetScale();
  for (SetIndex(0); InRange(); (*this)++)
  {
    Mapping().mapto(Quad()->p(GetIndex()), coord_quad());
    container_2d->surface_elev_quad()=coord_quad(2);
    for (int j=0; j<n_basis[1]; j++)
      prism->Basis(j)->basis_func_phys(coord_quad(), prism->Mapping(), basis_func_assign_quad(1, j));
  }
}


//----------------------------------------------------------------------------
//                        FACE_HORIZ_INTERIOR CLASS
//----------------------------------------------------------------------------

// Set recursively the face (or its descendants) to be active
// void FACE_HORIZ_INTERIOR::SetActiveRecursive()
// {
//   if (IsElement(0) && mesh.prism_list[Element(0)].IsActive() &&
//     IsElement(1) && mesh.prism_list[Element(1)].IsActive())
//   {
//     Status()|=ACTIVE;
//     UnSetUpdated();
//   }
//   else
//   {
//     Status()&=~ACTIVE;
//     for (int i=0; i<NChildren(); i++)
//       mesh.face_horiz_list[Child(i)].SetActiveRecursive();
//   }
// }

// Update rarely changed data at quadrature points (basis functions, coordinates, etc)
void FACE_HORIZ_INTERIOR::Update_basis_quad()
{
  const long_real direction_coef[2]={-1., 1.};
  PRISM_SW* prism=&mesh.prism_list[Element(0)];

  assert_debug(IsInitialized() && IsElement(0) && IsElement(1) && container_2d->IsUpdated());
  for (SetIndex(0); InRange(); (*this)++) // Update quadrature point coordinates
    Mapping().mapto(Quad()->p(GetIndex()), coord_quad());
  for (int k=0; k<2; k++)
  {
    PRISM_SW* prism=&mesh.prism_list[Element(k)];

    coef_scale(k)=direction_coef[k]*Mapping().GetScale()/prism->Mapping().GetScale();
    for (SetIndex(0); InRange(); (*this)++)
      for (int j=0; j<n_basis[k]; j++)  // Update basis functions
        prism->Basis(j)->basis_func_phys(coord_quad(), prism->Mapping(), basis_func_assign_quad(k, j));
  }
}



//--------------------------------------------------------------------------------------
//  FACE_VERT Class: Vertical face class for shallow-water problems
//--------------------------------------------------------------------------------------

// Default constructor
// Parameters: edge - corresponding 2D edge
FACE_VERT_SW::FACE_VERT_SW(EDGE_2D_SW& edge) : FACE_VERT(edge)
{
  container_2d=&edge;
  coef_scale_d[0]=coef_scale_d[1]=0.;
  n_basis[0]=n_basis[1]=0;
  for (int i=0; i<d_var; basis_index[0][i]=basis_index[1][i]=BAD_INDEX, i++);
  quad=quadrature_index.GetVertFaceQuad(problem_info.def_order);
}

// Constructor for hot start input
// Parameters: in - hot start input file stream
FACE_VERT_SW::FACE_VERT_SW(std::ifstream& in) : FACE_VERT(in)
{
  in.read((char*)coef_scale_d, sizeof coef_scale_d);
  in.read((char*)n_basis, sizeof n_basis);
  in.read((char*)basis_index, sizeof basis_index);
  container_2d=&mesh.edge_2d_list[Edge2D()];
  quad=quadrature_index.GetVertFaceQuad(problem_info.def_order);
}

// Compute mapping from the reference face
void FACE_VERT_SW::ComputeReferenceMapping()
{
  const real* coord[4];

  for (int i=0; i<4; coord[i]=mesh.node_3d_list[Node(i)].Coord(), i++);
  FACE_VERT::ComputeReferenceMapping(coord);
}

// Update recursively face-element connectivity info
// Parameters: el_ind - element index in the face; el_new - element id
void FACE_VERT_SW::SetElementRecursive(int el_ind, int el_new)
{
  assert_debug(el_ind>=0 && el_ind<2);
  Element(el_ind)=el_new;
  for (int i=0; i<NChildrenVert(); i++)
    mesh.face_vert_list[ChildVert(i)].SetElementRecursive(el_ind, el_new);
  for (int i=0; i<NChildrenHoriz(); i++)
    mesh.face_vert_list[ChildHoriz(i)].SetElementRecursive(el_ind, el_new);
}

// Set recursively the face (or its descendants) to be active
// This version works for boundary faces and must be overloaded for interior faces
void FACE_VERT_SW::SetActiveRecursive()
{
  bool active = IsInterior() ? IsElement(0) && IsElement(1) : true;

  for (int k=0; k<2; k++)
    active = IsElement(k) ? active && mesh.prism_list[Element(k)].IsActive() : active;
  if (active)
  {
    Status()|=ACTIVE;
    UnSetUpdated();
    for (int i=0; i<NChildrenHoriz(); i++)
      mesh.face_vert_list[ChildHoriz(i)].UnSetActiveRecursive();
    for (int i=0; i<NChildrenVert(); i++)
      mesh.face_vert_list[ChildVert(i)].UnSetActiveRecursive();
  }
  else
  {
    Status()&=~ACTIVE;
    for (int i=0; i<NChildrenHoriz(); i++)
      mesh.face_vert_list[ChildHoriz(i)].SetActiveRecursive();
    for (int i=0; i<NChildrenVert(); i++)
      mesh.face_vert_list[ChildVert(i)].SetActiveRecursive();
  }
}

// // Set recursively the face (or its descendants) to be active
// // This version works for boundary faces and must be overloaded for interior faces
// void FACE_VERT_SW::SetActiveRecursive()
// {
//   assert_debug(IsElement(0));
//   if (mesh.prism_list[Element(0)].IsActive())
//   {
//     Status()|=ACTIVE;
//     UnSetUpdated();
//     for (int i=0; i<NChildrenHoriz(); i++)
//       mesh.face_vert_list[ChildHoriz(i)].UnSetActiveRecursive();
//     for (int i=0; i<NChildrenVert(); i++)
//       mesh.face_vert_list[ChildVert(i)].UnSetActiveRecursive();
//   }
//   else
//   {
//     Status()&=~ACTIVE;
//     for (int i=0; i<NChildrenHoriz(); i++)
//       mesh.face_vert_list[ChildHoriz(i)].SetActiveRecursive();
//     for (int i=0; i<NChildrenVert(); i++)
//       mesh.face_vert_list[ChildVert(i)].SetActiveRecursive();
//   }
// }

// Set recursively the face (or its descendants) to be inactive
void FACE_VERT_SW::UnSetActiveRecursive()
{
  Status()&=~ACTIVE;
  for (int i=0; i<NChildrenVert(); i++)
    mesh.face_vert_list[ChildVert(i)].UnSetActiveRecursive();
  for (int i=0; i<NChildrenHoriz(); i++)
    mesh.face_vert_list[ChildHoriz(i)].UnSetActiveRecursive();
}

// Set recursively the face (or its descendants) to be updated
void FACE_VERT_SW::UnSetUpdatedRecursive()
{
  UnSetUpdated();
  for (int i=0; i<NChildrenVert(); i++)
    mesh.face_vert_list[ChildVert(i)].UnSetUpdatedRecursive();
  for (int i=0; i<NChildrenHoriz(); i++)
    mesh.face_vert_list[ChildHoriz(i)].UnSetUpdatedRecursive();
}

// Perform geometry update for moving free surface
void FACE_VERT_SW::UpdateSurfaceElevation()
{
  ComputeReferenceMapping();
  UnSetUpdated();
}

#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
// Parameters: el_ind - element index; k - 0=inflow, 1=outflow
// Return value: size of the inflow/outflow boundary, rvalue
real FACE_VERT_SW::JumpBoundary(int el_ind, int k) const
{
  assert_debug(el_ind>=0 && el_ind<2 && k>=0 && k<2 && IsElement(el_ind));
  return mesh.prism_list[Element(el_ind)].JumpBoundary(k);
}

// Parameters: el_ind - element index; k - 0=inflow, 1=outflow
// Return value: size of the inflow/outflow boundary, lvalue
real& FACE_VERT_SW::JumpBoundary(int el_ind, int k)
{
  assert_debug(el_ind>=0 && el_ind<2 && k>=0 && k<2 && IsElement(el_ind));
  return mesh.prism_list[Element(el_ind)].JumpBoundary(k);
}

// Parameters: el_ind - element index; k - 0=inflow, 1=outflow
// Return value: size of the jump in the unknown used for adaptivity over
//               the inflow/outflow boundary, rvalue
real FACE_VERT_SW::Jump(int el_ind, int k) const
{
  assert_debug(el_ind>=0 && el_ind<2 && k>=0 && k<2 && IsElement(el_ind));
  return mesh.prism_list[Element(el_ind)].Jump(k);
}

// Parameters: el_ind - element index; k - 0=inflow, 1=outflow
// Return value: size of the jump in the unknown used for adaptivity over
//               the inflow/outflow boundary, rvalue
real& FACE_VERT_SW::Jump(int el_ind, int k)
{
  assert_debug(el_ind>=0 && el_ind<2 && k>=0 && k<2 && IsElement(el_ind));
  return mesh.prism_list[Element(el_ind)].Jump(k);
}
#endif

#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
// Refine the face horizontally
// Parameters: node_new_ind - IDs of the new nodes, lvalue
// Return value: id of the first child face
void FACE_VERT_SW::RefineHoriz(int node_new_ind[2])
{
  if (!IsHorizRefined())
  {
    int n_children_horiz=mesh.edge_2d_list[Edge2D()].NChildren(),
      edge_child_2d=mesh.edge_2d_list[Edge2D()].Child(0);
    EDGE_2D *edge_2d_child[2];
    FACE_VERT *face_vert_child[2];
    NODE_3D *node_new[2], *node[4]={&mesh.node_3d_list[Node(0)], &mesh.node_3d_list[Node(1)], 
      &mesh.node_3d_list[Node(2)], &mesh.node_3d_list[Node(3)]};

    // Create new nodes
    for (int i=0; i<2; i++)
    {
      if (node_new_ind[i]==BAD_INDEX)
      {
        int node_2d_ind=mesh.edge_2d_list[edge_child_2d].Node(1);
        NODE_3D_SW node_loc(mesh.node_2d_list[node_2d_ind]);

        node_new_ind[i]=node_loc.ID()=mesh.node_3d_list.size();
        if (node[i]->IsSurface())
          node_loc.SetSurface();
        mesh.node_3d_list.push_back(node_loc);
      }
      node_new[i]=&mesh.node_3d_list[node_new_ind[i]];
    }
    // Create new faces
    for (int i=0; i<n_children_horiz; i++)
    {
      edge_2d_child[i]=&mesh.edge_2d_list[edge_child_2d+i];
      FACE_VERT_SW *face_vert_loc=CreateFaceVert(mesh.edge_2d_list[edge_child_2d+i]);

      face_vert_loc->ID()=mesh.face_vert_list.size();
      if (IsSurface())
        face_vert_loc->SetSurface();
      mesh.face_vert_list.push_back(*face_vert_loc);
      face_vert_child[i]=&mesh.face_vert_list[face_vert_loc->ID()];
      delete face_vert_loc;
    }
    FACE_VERT::RefineHoriz(edge_2d_child, face_vert_child, node, node_new);
    // Treat the case of vertically refined face
    if (IsVertRefined())
    {
      int node_ind_loc[2];

      for (int i=0; i<NChildrenVert(); i++)
      {
        if (i==0)
          node_ind_loc[0]=node_new_ind[0];
        else
          node_ind_loc[0]=node_ind_loc[1];
        if (i==NChildrenVert()-1)
          node_ind_loc[1]=node_new_ind[1];
        else
          node_ind_loc[1]=BAD_INDEX;
        mesh.face_vert_list[ChildVert(i)].RefineHoriz(node_ind_loc);
      }
      for (int i=0; i<n_children_horiz; i++)
      {
        FACE_VERT_SW& child_horiz=mesh.face_vert_list[ChildHoriz(i)];
        child_horiz.SetChildrenVert(NChildrenVert());
        for (int j=0; j<NChildrenVert(); j++)
          child_horiz.ChildVert(j)=mesh.face_vert_list[ChildVert(j)].ChildHoriz(i);
        child_horiz.Status()|=VERT_SPLIT;
        child_horiz.UnSetActive();
      }
    }
    for (int i=0; i<n_children_horiz; i++)
    {
      FACE_VERT_SW* face_child=(FACE_VERT_SW*)face_vert_child[i];

      face_child->ComputeReferenceMapping();
      face_child->InitializeStorage();
    }
  }
  else
  {
    for (int i=0; i<2; i++)
      node_new_ind[i]=mesh.face_vert_list[ChildHoriz(0)].Node(2+i);
  }
}
#endif

// Initialize storage for data stored at quadrature points
void FACE_VERT_SW::InitializeStorage()
{
  int n_quad=Quad()->GetNQuad();
  std::size_t ind=mass_storage_.size();

  BaseStorageFaceVert::Initialize(n_quad);
  for (int k=0; k<2; k++)
    if (IsElement(k))
    {
      PRISM_SW* prism=&mesh.prism_list[Element(k)];

      n_basis[k]=prism->NBasis();
      for (int i=0; i<d_var; i++)
        basis_index[k][i]=prism->Basis_index((VAR)i);
      for (int j=0; j<n_basis[k]; j++)
      {
        int n_dim=prism->Basis(j)->Dim();

        mass_storage_.insert(mass_storage_.end(), n_quad*n_dim, 0);
        for (SetIndex(0); InRange(); ind+=n_dim, (*this)++)
          quad_local().basis_func_ind[k][j]=ind;
      }
    }
}

// Update data shared by all vertical faces
// Parameters: var_new - global storage for degrees of freedom;
//             var_flux_new - global storage for degrees of freedom for the fluxes;
//             var_rhs_new - global storage for degrees of freedom for the right hand side
void FACE_VERT_SW::Update_static(DOF var_new[d_var], DOF var_flux_new[d_var][d_space], 
                                 DOF var_rhs_new[d_var])
{
  for (int i=0; i<d_var; i++)
  {
    temp_storage.Var_dof[i]=var_new[i].GetDofIterator();
    temp_storage.L_Var_dof[i]=var_rhs_new[i].GetDofIterator();
    for (int j=0; j<d_space; j++)
      temp_storage.Q_dof[i][j]=var_flux_new[i][j].GetDofIterator();        
  }
}

// Update rarely changed data at quadrature points (basis functions, coordinates, etc)
void FACE_VERT_SW::Update_basis_quad()
{
  const long_real direction_coef[2]={1., -1.};
  assert_debug(IsInitialized() && container_2d->IsUpdated());

  for (SetIndex(0); InRange(); (*this)++) // Update coordinates and weights
  {
    Mapping().mapto(Quad()->p(GetIndex()), coord_quad());
    weight_quad()=Quad()->w(GetIndex())*Mapping().GetCorrection(Quad()->p(GetIndex()));
  }
  for (int k=0; k<2; k++)
    if(IsElement(k))
    {
      PRISM_SW* prism=&mesh.prism_list[Element(k)];

      coef_scale(k)=direction_coef[k]/prism->Mapping().GetScale();
      for (SetIndex(0); InRange(); (*this)++) // Update basis functions
        for (int j=0; j<n_basis[k]; j++)
          prism->Basis(j)->basis_func_phys(coord_quad(), prism->Mapping(), basis_func_assign_quad(k, j));
    }
}

// // Update often changed data at quadrature points (unknowns)
// // Parameters: t - time
// void FACE_VERT_SW::Update_variables_quad(real t)
// {
//   static const VAR* ind=VAR_3D_DT_INDEX;
//   int boundary_ind=-1;

//   assert_debug(IsUpdated());
//   for (int k=0; k<2; k++)
//     if (IsElement(k))      
//       for (SetIndex(0); InRange(); (*this)++)
//         for (int i=0; i<d_var_3d_dt; i++)
//           quad_local().Var[k][ind[i]]=Var_dof(k, ind[i])*basis_func_quad(k, ind[i]);
//     else
//       boundary_ind=k;
//   if (boundary_ind==-1)
//     return;
//   for (SetIndex(0); InRange(); (*this)++)
//     BC_Dirichlet_quad(t, 0, quad_local().Var[boundary_ind]);
// }

#ifdef BAROCLINIC_KEY
// Parameters: el_ind - element index on the face
// Return value: pointer to values of vertically integrated basis functions
//               at the quadrature point, rvalue
const real* FACE_VERT_SW::basis_func_integrated_quad(int el_ind) const
{
  assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind));
  static func_type basis_func_integrated;
  PRISM_SW* prism=&mesh.prism_list[Element(el_ind)];

  prism->Basis_dof(RHO)->
    basis_func_vert_integral_phys(coord_quad(), prism->Mapping(), basis_func_integrated);
  return basis_func_integrated;
}
#endif

// Hot start output
std::ofstream& operator<<(std::ofstream& out, const FACE_VERT_SW& face)
{
  out << (FACE_VERT)face;
  out.write((char*)face.coef_scale_d, sizeof face.coef_scale_d);
  out.write((char*)face.n_basis, sizeof face.n_basis);
  out.write((char*)face.basis_index, sizeof face.basis_index);
  return out;
}



//--------------------------------------------------------------------------------------
//  FACE_VERT_INTERIOR Class: Vertical interior face class for shallow-water problems
//--------------------------------------------------------------------------------------

// // Set recursively the face (or its descendants) to be active
// void FACE_VERT_INTERIOR::SetActiveRecursive()
// {
//   if (IsElement(0) && mesh.prism_list[Element(0)].IsActive() &&
//       IsElement(1) && mesh.prism_list[Element(1)].IsActive())
//   {
//     Status()|=ACTIVE;
//     UnSetUpdated();
//     for (int i=0; i<NChildrenHoriz(); i++)
//       mesh.face_vert_list[ChildHoriz(i)].UnSetActiveRecursive();
//     for (int i=0; i<NChildrenVert(); i++)
//       mesh.face_vert_list[ChildVert(i)].UnSetActiveRecursive();
//   }
//   else
//   {
//     Status()&=~ACTIVE;
//     for (int i=0; i<NChildrenHoriz(); i++)
//       mesh.face_vert_list[ChildHoriz(i)].SetActiveRecursive();
//     for (int i=0; i<NChildrenVert(); i++)
//       mesh.face_vert_list[ChildVert(i)].SetActiveRecursive();
//   }
// }



//--------------------------------------------------------------------------------------
// PRISM_SW Class: Element class in 3D for shallow-water problems
//--------------------------------------------------------------------------------------

// Default constructor
// Parameters: elem_2d - corresponding 2D element
PRISM_SW::PRISM_SW(TRIANGLE_2D_SW& elem_2d) : PRISM(elem_2d)
{
  Type()=PARALLEL;
  container_2d=&elem_2d;
  order=problem_info.def_order;
  quad=quadrature_index.GetPrismQuad(order);
  vert=quadrature_index.GetPrismNodeQuad();
  w_matrix_ind=(std::size_t)BAD_INDEX;
  n_basis=0;
  for (int i=0; i<d_var_3d; i++)
    basis[i]=0;
  for (int i=0; i<d_var; basis_index[i]=BAD_INDEX, i++);
#ifdef ADAPTIVITY_H_3D_KEY
  min=max=abs_max=s_jump[0]=s_jump[1]=s_boundary[0]=s_boundary[1]=0.;
#endif
}

// Constructor for hot start input
// Parameters: in - hot start input file stream
PRISM_SW::PRISM_SW(std::ifstream& in) : PRISM(in)
{
  in.read((char*)&order, sizeof order);
  in.read((char*)&n_basis, sizeof n_basis);
  in.read((char*)basis_index, sizeof basis_index);
  container_2d=&mesh.triangle_2d_list[Element2D()];
  quad=quadrature_index.GetPrismQuad(order);
  vert=quadrature_index.GetPrismNodeQuad();
  w_matrix_ind=(std::size_t)BAD_INDEX;
  for (int i=0; i<d_var_3d; i++)
    basis[i]=0;
#ifdef ADAPTIVITY_H_3D_KEY
  in.read((char*)&min, sizeof min);
  in.read((char*)&max, sizeof max);
  in.read((char*)s_jump, sizeof s_jump);
  in.read((char*)s_boundary, sizeof s_boundary);
#endif
}

// Compute mapping from the reference triangle
void PRISM_SW::ComputeReferenceMapping()
{
  const real* coord[2*vert_h];
  FACE_HORIZ_SW &top_face=mesh.face_horiz_list[TopFace()], &bot_face=mesh.face_horiz_list[BottomFace()];

  for (int i=0; i<vert_h; i++)
    coord[i]=mesh.node_3d_list[top_face.Node(i)].Coord(),
    coord[vert_h+i]=mesh.node_3d_list[bot_face.Node(i)].Coord();
  PRISM::ComputeReferenceMapping(coord);
}

// Perform geometry update for moving free surface
void PRISM_SW::UpdateSurfaceElevation()
{
  // Update geometry and quadratures
  ComputeReferenceMapping();
  UnSetUpdated();
  int w_dim=Basis_dof(W)->Dim();
  std::memset(w_matrix(), 0, w_dim*w_dim*sizeof(*w_matrix()));
}


#ifdef ADAPTIVITY_H_3D_KEY
// Initialize data used for shock-detector
void PRISM_SW::InitializeAdaptivityData()
{
  for (int k=0; k<2; k++)
    s_jump[k]=s_boundary[k]=0.;
  // Compute minimum/maximum values of the adaptivity variable
  min=1.e25;
  max=-1.e25;  
  for (SetIndex(0); InRange(); (*this)++)
  {
    real z_s=Var_quad(problem_info.refinement_var);
    min = min > z_s ? z_s : min;
    max = max < z_s ? z_s : max;
  }
  abs_max = max>-min ? max : min;
}
#endif

#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
// Refine the element by splitting in four
// Return value: id of the first child element
void PRISM_SW::Refine()
{
  if (!IsRefined())
  {
    int node_new_ind[2][vert_h]={{BAD_INDEX, BAD_INDEX, BAD_INDEX},
    {BAD_INDEX, BAD_INDEX, BAD_INDEX}}; 
    int face_top_child_ind, face_bot_child_ind,
      n_children=mesh.triangle_2d_list[Element2D()].NChildren(),
      element_2d_child_ind=mesh.triangle_2d_list[Element2D()].Child(0);
    TRIANGLE_2D *element_2d_child[4]={&mesh.triangle_2d_list[element_2d_child_ind],
      &mesh.triangle_2d_list[element_2d_child_ind+1],
      &mesh.triangle_2d_list[element_2d_child_ind+2],
      &mesh.triangle_2d_list[element_2d_child_ind+3]};
    EDGE_2D *edge_2d_new[vert_h];
    NODE_3D *node_new[2][vert_h];
    FACE_HORIZ *face_top_child[4], *face_bot_child[4];
    FACE_VERT *face_vert_child[vert_h][2], *face_vert_new[vert_h];
    PRISM *prism_child[4];

    // Get existing nodes
    if (mesh.face_horiz_list[TopFace()].IsRefined())
      mesh.face_horiz_list[TopFace()].Refine(node_new_ind[0]);
    if (mesh.face_horiz_list[BottomFace()].IsRefined())
      mesh.face_horiz_list[BottomFace()].Refine(node_new_ind[1]);

    // Refine vertical faces and create new vertical faces and nodes
    for (int i=0; i<vert_h; i++)
    {
      int node_new_vert_ind[2]={node_new_ind[0][i], node_new_ind[1][i]};

      mesh.face_vert_list[SideFace(i)].RefineHoriz(node_new_vert_ind);
      edge_2d_new[i]=&mesh.edge_2d_list[element_2d_child[n_children-1]->Edge(i)];
      FACE_VERT_INTERIOR face_vert_new_loc((EDGE_INTERIOR&)*edge_2d_new[i]);
      face_vert_new_loc.ID()=mesh.face_vert_list.size();
      if (mesh.face_horiz_list[TopFace()].IsSurface())
        face_vert_new_loc.SetSurface();
      mesh.face_vert_list.push_back(face_vert_new_loc);
      node_new_ind[0][i]=node_new_vert_ind[0];
      node_new[0][i]=&mesh.node_3d_list[node_new_vert_ind[0]];
      node_new_ind[1][i]=node_new_vert_ind[1];
      node_new[1][i]=&mesh.node_3d_list[node_new_vert_ind[1]];
      face_vert_child[i][0]=&mesh.face_vert_list[mesh.face_vert_list[SideFace(i)].ChildHoriz(0)];
      face_vert_child[i][1]=&mesh.face_vert_list[mesh.face_vert_list[SideFace(i)].ChildHoriz(1)];
      face_vert_new[i]=&mesh.face_vert_list[face_vert_new_loc.ID()];
    }
    // Refine horizontal faces
    if (!mesh.face_horiz_list[TopFace()].IsRefined())
      mesh.face_horiz_list[TopFace()].Refine(node_new_ind[0]);
    if (!mesh.face_horiz_list[BottomFace()].IsRefined())
      mesh.face_horiz_list[BottomFace()].Refine(node_new_ind[1]);
    face_top_child_ind=mesh.face_horiz_list[TopFace()].Child(0);
    face_bot_child_ind=mesh.face_horiz_list[BottomFace()].Child(0);

    // Create new prisms
    for (int i=0; i<n_children; i++)
    {
      PRISM_SW* prism_child_loc=CreatePrism(mesh.triangle_2d_list[element_2d_child_ind+i]);

      prism_child_loc->ID()=mesh.prism_list.size();
      prism_child_loc->Type()=Type();
      mesh.prism_list.push_back(*prism_child_loc);
      prism_child[i]=&mesh.prism_list[prism_child_loc->ID()];
      delete prism_child_loc;
      face_top_child[i]=&mesh.face_horiz_list[face_top_child_ind+i];
      face_bot_child[i]=&mesh.face_horiz_list[face_bot_child_ind+i];
    }
    PRISM::Refine(element_2d_child, edge_2d_new, prism_child, face_vert_new, face_vert_child, 
      face_top_child, face_bot_child, node_new);
    // Update prisms
    for (int i=0; i<n_children; i++)
    {
      PRISM_SW& prism_child_loc=(PRISM_SW&)*prism_child[i];

      prism_child_loc.ComputeReferenceMapping();
      for (int j=0; j<d_var_3d; j++)
        (Var_dof(VAR_3D_INDEX[j]).GetDof())->
          InitializeLocal(prism_child_loc.ID(), Var_dof(VAR_3D_INDEX[j]).GetDim());
      prism_child_loc.InitializeStorage();
//         var[VAR_3D_INDEX[j]].
//         InitializeLocal(prism_child_loc.ID(), Var_dof(VAR_3D_INDEX[j]).GetDim());
    }
    // Update new vertical faces
    for (int i=0; i<vert_h; i++)
    {
      ((FACE_VERT_INTERIOR*)face_vert_new[i])->ComputeReferenceMapping();
      ((FACE_VERT_INTERIOR*)face_vert_new[i])->InitializeStorage();
    }
  }

  UnSetActive();
  mesh.DecrementActiveElementCnt();
  mesh.face_horiz_list[TopFace()].UnSetActiveRecursive();
  mesh.face_horiz_list[BottomFace()].UnSetActiveRecursive();
  for (int i=0; i<vert_h; i++)
    mesh.face_vert_list[SideFace(i)].UnSetActiveRecursive();

  for (int i=0; i<NChildren(); i++)
  {
    PRISM_SW& prism_child=mesh.prism_list[Child(i)];
    FACE_HORIZ_SW &top=mesh.face_horiz_list[prism_child.TopFace()], 
      &bot=mesh.face_horiz_list[prism_child.BottomFace()];

    // Update connectivity information
    prism_child.SetActive();
    mesh.IncrementActiveElementCnt();
    prism_child.ComputeReferenceMapping();
    prism_child.Update_basis_quad();
    top.SetElementRecursive(1, Child(i));
    top.SetActiveRecursive();
    bot.SetElementRecursive(0, Child(i));
    bot.SetActiveRecursive();
    for (int j=0; j<vert_h; j++)
    {
      FACE_VERT_SW &side=mesh.face_vert_list[prism_child.SideFace(j)];

      side.SetElementRecursive(prism_child.ElementIndex(j), Child(i));
      side.SetActiveRecursive();

    }

    // Project variables to the child element
    for (int j=0; j<d_var_3d; j++)
      prism_child.Var_dof(VAR_3D_INDEX[j])=0.;
    for (prism_child.SetIndex(0); prism_child.InRange(); prism_child++)
    {
      func_type basis_func[d_var_3d];
      for (int j=0; j<n_basis; j++)
        Basis(j)->basis_func_phys(prism_child.coord_quad(), Mapping(), basis_func[j]);
      for (int j=0; j<d_var_3d; j++)
      {
        VAR ind=VAR_3D_INDEX[j];
        long_real var_tmp=Var_dof(ind)*basis_func[Basis_index(ind)];
        prism_child.Var_dof(ind).
          Add(prism_child.weight_quad()*var_tmp, prism_child.basis_func_quad(ind));
      }
    }
  }      
}

// Unrefine the element
void PRISM_SW::UnRefine()
{
  assert_debug(NChildren()>0);

  SetActive();
  mesh.IncrementActiveElementCnt();
  ComputeReferenceMapping();
  Update_basis_quad();
  for (int i=0; i<d_var_3d; i++)
    Var_dof(VAR_3D_INDEX[i])=0.;
  for (int i=0; i<NChildren(); i++)
  {
    PRISM_SW& prism_child=mesh.prism_list[Child(i)];
    func_type basis_func[d_var_3d];

    prism_child.UnSetActive();
    mesh.DecrementActiveElementCnt();
    mesh.face_horiz_list[prism_child.TopFace()].UnSetActiveRecursive();
    mesh.face_horiz_list[prism_child.BottomFace()].UnSetActiveRecursive();
    for (int j=0; j<vert_h; j++)
      mesh.face_vert_list[prism_child.SideFace(j)].UnSetActiveRecursive();
    // Project variables to the parent element
    for (SetIndex(0); InRange(); (*this)++)
    {
      if (prism_child.container_2d->DistanceToBoundary(coord_quad())>0.)
        continue;
      for (int j=0; j<n_basis; j++)
        prism_child.Basis(j)->basis_func_phys(coord_quad(), prism_child.Mapping(), basis_func[j]);
      for (int j=0; j<d_var_3d; j++)
      {
        VAR ind=VAR_3D_INDEX[j];
        long_real var_tmp=prism_child.Var_dof(ind)*basis_func[Basis_index(ind)];
        Var_dof(ind).Add(weight_quad()*var_tmp, basis_func_quad(ind));
      }
    }
  }
  mesh.face_horiz_list[TopFace()].SetElementRecursive(1, ID());
  mesh.face_horiz_list[TopFace()].SetActiveRecursive();
  mesh.face_horiz_list[BottomFace()].SetElementRecursive(0, ID());
  mesh.face_horiz_list[BottomFace()].SetActiveRecursive();
  for (int j=0; j<vert_h; j++)
  {
    mesh.face_vert_list[SideFace(j)].SetElementRecursive(ElementIndex(j), ID());
    mesh.face_vert_list[SideFace(j)].SetActiveRecursive();
  }
}
#endif

// Initialize storage for data stored at quadrature points
void PRISM_SW::InitializeStorage()
{
  int n_quad=Quad()->GetNQuad();
  BaseStoragePrism::Initialize(n_quad);
  std::size_t ind=mass_storage_.size();

  // Initialize matrix for w computation
  int w_dim=Basis_dof(W)->Dim(), length_w=w_dim*w_dim;
  w_matrix_ind=ind;
  mass_storage_.insert(mass_storage_.end(), length_w, 0);
  ind+=length_w;

  for (int j=0; j<n_basis; j++)
  {
    int n_dim=basis[j]->Dim();

    mass_storage_.insert(mass_storage_.end(), n_quad*d_space*n_dim, 0);
    for (SetIndex(0); InRange(); (*this)++)
      for (int k=0; k<d_space; k++)
        quad_local().basis_grad_ind[j][k]=ind, ind+=n_dim;
  }
}

// Update data shared by all prisms
// Parameters: quad - pointer to quadrature object;
//             basis - pointer to basis object
//             var_new - global storage for degrees of freedom;
//             var_flux_new - global storage for degrees of freedom for the fluxes;
//             var_rhs_new - global storage for degrees of freedom for the right hand side
void PRISM_SW::Update_static(const QUADRATURE *quad, const BASIS *basis[max_order+1], DOF var_new[d_var], 
                             DOF var_flux_new[d_var][d_space], DOF var_rhs_new[d_var])
{
  static const VAR* ind=VAR_3D_INDEX;
  int i, j, n_basis=0, n_quad=quad->GetNQuad();

  // Initialize pointers to basis objects
  for (i=0; i<d_var_3d; i++)
  {
    for (j=0; j<i; j++)
      if (problem_info.var_order[ind[i]]==problem_info.var_order[ind[j]])
        break;
    if (i>0 && j<i)
      continue;
    // Assign new basis storage
    int order=problem_info.var_order[ind[i]], n_dim=basis[order]->Dim(), ind=mass_storage_.size();
    mass_storage_.insert(mass_storage_.end(), n_quad*n_dim, 0);
    for (j=0; j<n_quad; ind+=n_dim, j++)
    {
      quad_temp_storage[j].basis_func_ind[n_basis]=ind;
      basis[order]->basis_func_ref(quad->p(j), &mass_storage_[ind]);
    }
    n_basis++;
  }
//   for (int j=0; j<=max_order; j++)
//     if (basis[j]!=0) // Storage for basis functions
//     {
//       PRISM_SW* tmp_prism=0;
//       int n_dim=basis[j]->Dim(), ind=mass_storage_.size(), 
//         length_basis=n_dim*sizeof(*(tmp_prism->basis_func_assign_quad(j)))/sizeof(mass_storage_[0]);
//       mass_storage_.insert(mass_storage_.end(), n_quad*length_basis, 0);
//       for (int i=0; i<n_quad; ind+=length_basis, i++)
//       {
//         quad_temp_storage[i].basis_func_ind[j]=ind;
//         basis[j]->basis_func_ref(quad->p(i), &mass_storage_[ind]);
//       }
//     }
  for (i=0; i<d_var; i++)
  {
    temp_storage.Var_dof[i]=var_new[i].GetDofIterator();
    temp_storage.L_Var_dof[i]=var_rhs_new[i].GetDofIterator();
    for (j=0; j<d_space; j++)
      temp_storage.Q_dof[i][j]=var_flux_new[i][j].GetDofIterator();        
  }
}

// Update rarely changed data at quadrature points (basis functions, coordinates, etc)
void PRISM_SW::Update_basis_quad()
{
  assert_debug(IsInitialized() && container_2d->IsUpdated());
  for (SetIndex(0); InRange(); (*this)++)
  {
    const real *xi=quad->p(GetIndex());

    Mapping().mapto(xi, coord_quad());
    weight_quad()=Quad()->w(GetIndex())*Mapping().GetCorrection(xi);
    for (int i=0; i<n_basis; i++)
    {
      int n_dim=Basis(i)->Dim();
      grad_type basis_grad_tmp;

      Basis(i)->basis_grad_ref(xi, Mapping(), basis_grad_tmp);
      for (int k=0; k<n_dim; k++)
        for (int j=0; j<d_space; j++)
          basis_grad_assign_quad(i, j)[k]=basis_grad_tmp[k][j];
    }
  }
}

// Output in binary format
std::ofstream& operator<<(std::ofstream& out, const PRISM_SW& element)
{
  out << (PRISM)element;
  out.write((char*)&element.order, sizeof element.order);
  out.write((char*)&element.n_basis, sizeof element.n_basis);
  out.write((char*)element.basis_index, sizeof element.basis_index);
#ifdef ADAPTIVITY_H_3D_KEY
  out.write((char*)&element.min, sizeof element.min);
  out.write((char*)&element.max, sizeof element.max);
  out.write((char*)element.s_jump, sizeof element.s_jump);
  out.write((char*)element.s_boundary, sizeof element.s_boundary);
#endif
  return out;
}


//--------------------------------------------------------------------------------------
// PRISM_PARALLEL Class: Parallel prism class in 3D for shallow-water problems
//--------------------------------------------------------------------------------------

// Default constructor
// Parameters: elem_2d - corresponding 2D element
PRISM_PARALLEL::PRISM_PARALLEL(TRIANGLE_2D_SW& face) : PRISM_SW(face)
{
  static const VAR* ind=VAR_3D_INDEX;

  Type()=PARALLEL;
  // Initialize pointers to basis objects
  for (int i=0; i<d_var_3d; i++)
  {
    for (int j=0; j<i; j++)
      if (problem_info.var_order[ind[i]]==problem_info.var_order[ind[j]])
      {
        basis_index[ind[i]]=basis_index[ind[j]];
        break;
      }
    // Assign new basis object
    if (basis_index[ind[i]]==BAD_INDEX)
    {
      basis[n_basis]=basis_ref[problem_info.var_order[ind[i]]];
      basis_index[ind[i]]=n_basis++;
    }
  }
//   n_basis=Order()+1;
//   for (int i=0; i<n_basis; i++)
//     basis[i] = basis_ref[Order()-i];
//   for (int i=0; i<d_var; i++)
//   {
//     if (VAR_3D_INDEX_INV[i]==BAD_INDEX)
//     {
//       basis_index[i]=BAD_INDEX;
//       continue;
//     }
//     basis_index[i]=VAR_BASIS_INDEX[i];
//   }
}

// Constructor for hot start input
// Parameters: in - hot start input file stream
PRISM_PARALLEL::PRISM_PARALLEL(std::ifstream& in) : PRISM_SW(in)
{
  static const VAR* ind=VAR_3D_INDEX;

  // Initialize pointers to basis objects
  for (int i=0; i<d_var_3d; i++)
  {
    assert_debug(basis_index[ind[i]]!=BAD_INDEX);
    basis[basis_index[ind[i]]]=basis_ref[problem_info.var_order[ind[i]]];
  }
//   n_basis=Order()+1;
//   for (int i=0; i<n_basis; i++)
//     basis[i] = basis_ref[Order()-i];
}

// Initialize storage for data stored at quadrature points
void PRISM_PARALLEL::InitializeStorage()
{
  PRISM_SW::InitializeStorage();

  for (SetIndex(0); InRange(); (*this)++)
    for (int j=0; j<n_basis; j++)
      quad_local().basis_func_ind[j]=quad_temp().basis_func_ind[j];
}



//--------------------------------------------------------------------------------------
// PRISM_NONPARALLEL Class: Nonparallel prism class in 3D for shallow-water problems
//--------------------------------------------------------------------------------------

// Default constructor
// Parameters: elem_2d - corresponding 2D element
PRISM_NONPARALLEL::PRISM_NONPARALLEL(TRIANGLE_2D_SW& elem_2d) : PRISM_SW(elem_2d)
{
  Type()=NONPARALLEL;
  for (int i=0; i<d_var_3d; mass_matrix_ind[i]=(std::size_t)BAD_INDEX, i++);
}

// Constructor for hot start input
// Parameters: in - hot start input file stream
PRISM_NONPARALLEL::PRISM_NONPARALLEL(std::ifstream& in) : PRISM_SW(in)
{
  static const VAR* ind=VAR_3D_INDEX;

  for (int i=0; i<d_var_3d; i++)
    mass_matrix_ind[i]=(std::size_t)BAD_INDEX;

  // Initialize pointers to basis objects
  for (int i=0; i<d_var_3d; i++)
  {
    assert_debug(basis_index[ind[i]]!=BAD_INDEX);
    if (basis[basis_index[ind[i]]]!=0)
      continue;
    // Create new basis object
    switch(problem_info.var_order[ind[i]])
    {
    case 0:
      basis_phys_const.push_back(BASIS_ZPHYS_CONST(in));
      basis[basis_index[ind[i]]]=&basis_phys_const[basis_phys_const.size()-1];
      break;
    case 1:
      basis_phys_lin.push_back(BASIS_ZPHYS_LIN(in));
      basis[basis_index[ind[i]]]=&basis_phys_lin[basis_phys_lin.size()-1];
      break;
    case 2:
      basis_phys_quad.push_back(BASIS_ZPHYS_QUAD(in));
      basis[basis_index[ind[i]]]=&basis_phys_quad[basis_phys_quad.size()-1];
      break;
    default: std::cerr << "File mesh_3d_sw.cpp: Wrong basis choice parameter!\n";
    }
  }
//   for (int i=0; i<n_basis; i++)
//     switch(Order()-i)
//     {
//     case 0:  // Default basis
//       basis_phys_const.push_back(BASIS_ZPHYS_CONST(in));
//       basis[i]=&basis_phys_const[basis_phys_const.size()-1];
//       break;
//     case 1:
//       basis_phys_lin.push_back(BASIS_ZPHYS_LIN(in));
//       basis[i]=&basis_phys_lin[basis_phys_lin.size()-1];
//       break;
//     case 2:
//       basis_phys_quad.push_back(BASIS_ZPHYS_QUAD(in));
//       basis[i]=&basis_phys_quad[basis_phys_quad.size()-1];
//       break;
//     default: std::cerr << "File mesh_3d_sw.cpp: Wrong basis choice parameter!\n";
//     }
}

// Compute mapping from the reference triangle
void PRISM_NONPARALLEL::ComputeReferenceMapping()
{
  static const VAR* ind=VAR_3D_INDEX;

  PRISM_SW::ComputeReferenceMapping();
  if (n_basis!=0)
    return;
  // Initialize pointers to basis objects
  for (int i=0; i<d_var_3d; i++)
  {
    for (int j=0; j<i; j++)
      if (problem_info.var_order[ind[i]]==problem_info.var_order[ind[j]])
      {
        basis_index[ind[i]]=basis_index[ind[j]];
        break;
      }
    if (basis_index[ind[i]]!=BAD_INDEX)
      continue;
    // Create new basis object
    switch(problem_info.var_order[ind[i]])
    {
    case 0:
      basis_phys_const.push_back(BASIS_ZPHYS_CONST(Mapping()));
      basis[n_basis]=&basis_phys_const[basis_phys_const.size()-1];
      break;
    case 1:
      basis_phys_lin.push_back(BASIS_ZPHYS_LIN(Mapping()));
      basis[n_basis]=&basis_phys_lin[basis_phys_lin.size()-1];
      break;
    case 2:
      basis_phys_quad.push_back(BASIS_ZPHYS_QUAD(Mapping()));
      basis[n_basis]=&basis_phys_quad[basis_phys_quad.size()-1];
      break;
    default: std::cerr << "File mesh_3d_sw.cpp: Wrong basis choice parameter!\n";
    }
    basis_index[ind[i]]=n_basis++;
  }
//     n_basis=Order()+1;
//     for (int i=0; i<n_basis; i++)
//       switch(Order()-i)
//       {
//       case 0:  // Default basis
//         basis_phys_const.push_back(BASIS_ZPHYS_CONST(Mapping()));
//         basis[i]=&basis_phys_const[basis_phys_const.size()-1];
//         break;
//       case 1:
//         basis_phys_lin.push_back(BASIS_ZPHYS_LIN(Mapping()));
//         basis[i]=&basis_phys_lin[basis_phys_lin.size()-1];
//         break;
//       case 2:
//         basis_phys_quad.push_back(BASIS_ZPHYS_QUAD(Mapping()));
//         basis[i]=&basis_phys_quad[basis_phys_quad.size()-1];
//         break;
//       default: std::cerr << "File mesh_3d_sw.cpp: Wrong basis choice parameter!\n";
//       }
//     for (int i=0; i<d_var; i++)
//     {
//       if (VAR_3D_INDEX_INV[i]==BAD_INDEX)
//       {
//         basis_index[i]=BAD_INDEX;
//         continue;
//       }
//       basis_index[i]=VAR_BASIS_INDEX[i];
//     }
}

// Invert mass matrices for all basis functions
void PRISM_NONPARALLEL::InvertMassMatrix()
{
  for (int i=0; i<n_basis; i++)
  {
    int n_dim=Basis(i)->Dim();

    std::memset(mass_matrix_assign(i), 0, n_dim*n_dim*sizeof(*mass_matrix_assign(i)));       
    for (SetIndex(0); InRange(); (*this)++)
      for (int j=0, ind=0; j<n_dim; j++)
        for (int k=0; k<n_dim; ind++, k++)
          mass_matrix_assign(i)[ind]+=
            weight_quad()*basis_func_assign_quad(i)[k]*basis_func_assign_quad(i)[j];
    if (LU_decomposition(n_dim, mass_matrix_assign(i))==BAD_INDEX)
    {
      std::cerr << "File mesh_3d_sw.cpp: Singular mass matrix!\n";
      std::exit(1);
    }
  }
}

#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
// Refine the element
// Return value: id of the first child element
void PRISM_NONPARALLEL::Refine()
{
  PRISM_SW::Refine();
  for (int i=0; i<NChildren(); i++)
  {
    PRISM_NONPARALLEL& prism_child=(PRISM_NONPARALLEL&)mesh.prism_list[Child(i)];

    // Project variables to the sub-element
    for (int k=0; k<d_var_3d; k++)
      LU_substitution(prism_child.Var_dof(VAR_3D_INDEX[k]).GetDim(),
                      prism_child.mass_matrix(VAR_3D_INDEX[k]),
                      &prism_child.Var_dof(VAR_3D_INDEX[k]));
  }      
}

// Unrefine the element
void PRISM_NONPARALLEL::UnRefine()
{
  PRISM_SW::UnRefine();
  for (int k=0; k<d_var_3d; k++)
    LU_substitution(Var_dof(VAR_3D_INDEX[k]).GetDim(), mass_matrix(VAR_3D_INDEX[k]),
                    &Var_dof(VAR_3D_INDEX[k]));
}
#endif

// Initialize storage for data stored at quadrature points
void PRISM_NONPARALLEL::InitializeStorage()
{
  PRISM_SW::InitializeStorage();
  int n_quad=Quad()->GetNQuad();
  std::size_t ind=mass_storage_.size();

  for (int j=0; j<n_basis; j++)
  {
    int n_dim=basis[j]->Dim(), length_mass=n_dim*n_dim;

    // Storage for basis functions
    mass_storage_.insert(mass_storage_.end(), n_quad*n_dim, 0);
    for (SetIndex(0); InRange(); ind+=n_dim, (*this)++)
      quad_local().basis_func_ind[j]=ind;

    // Storage for the mass matrix
    mass_storage_.insert(mass_storage_.end(), length_mass, 0);
    mass_matrix_ind[j]=ind;
    ind+=length_mass;
  }
}

// Update rarely changed data at quadrature points (basis functions, coordinates, etc)
void PRISM_NONPARALLEL::Update_basis_quad()
{
  PRISM_SW::Update_basis_quad();
  for (SetIndex(0); InRange(); (*this)++)
    for (int i=0; i<n_basis; i++)
      Basis(i)->basis_func_phys(coord_quad(), Mapping(), basis_func_assign_quad(i));
  InvertMassMatrix();
}

// Hot start output
std::ofstream& operator<<(std::ofstream& out, const PRISM_NONPARALLEL& element)
{
  out << (PRISM_SW)element;
  for (int i=0; i<element.NBasis(); i++)
    out << (BASIS_ZPHYS&)*element.Basis(i);
  return out;
}
