//----------------------------------------------------------------------------------------
// pde.H: Declaration for the class used for storage of physical parameters
//----------------------------------------------------------------------------------------

#ifndef PDE_H
#define PDE_H

#include <iostream>
#include <vector>
#include <cmath>
#include "general.H"

// Storage for tidal forcing parameters for one frequency
typedef struct tp_info
{
  char tp_tag[6];   // Tag for the tidal potential frequency
  real tp_coef;     // Tidal potential coefficient
  real amig;        // Frequency
  real etrf;        // Earth tide potential reduction factor
  real ff;          // Time dependent nodal factor
  real phase;       // Initial phase
} tide_info;

// Storage for shallow-water problem parameters
struct PROBLEM_DESCRIPTION
{
  enum coord_sys {CARTESIAN, SPHERICAL};
  enum const_variable {CONSTANT, VARIABLE};

  // Physical constants
  long_real Earth_radius;      // Radius of the Earth in meters
  long_real pi;                // Pi number
  int n_layers;                // Number of vertical layers
  int mesh_update;             // Mesh update frequency (0 - never, n - every n time steps)
  real mesh_grading_bot;       // Mesh grading at the bottom parameter (0-no restrictions,=dz_bottom/dz) 
  int runge_kutta_method;      // 1 = Forward Euler, 2 = 2-stage Runge-Kutta, 3 = 3-stage Runge-Kutta
  int max_refinement;          // Maximum level of refinement for h-adaptivity
  VAR refinement_var;          // Variable used for shock detector
  real refinement_start;       // Refinement start time for h-adaptivity
  real detector_sensitivity;   // Shock detector sensitivity
  int def_order;               // Default order of approximating spaces
  int var_order[d_var];        // Order of approximating space for each variable
  char output_file[file_name]; // Common name for ouput files
  char tecplot_out[file_name]; // Name of Tecplot ouput file
  char elev_out[file_name];    // Name of elevation recording stations ouput file
  char vel_out[file_name];     // Name of velocity recording stations ouput file
  char run_description[32];    // Run description
  char run_id[24];             // Run ID
  char grid_desc[24];          // Grid descriptor
  int hot_input;               // Hot/cold start option parameter switch
  int hot_output;              // Hot output control option
  char hot_input_file[file_name];  // Name of the hot input file
  int hot_freq;                // Hot output control frequency (every n time steps)
  coord_sys coord_system;      // Type of coordinate system switch
  const_variable coriolis_par; // Const/variable Coriolis parameter switch
  bool tidal_poten;            // Tidal potential switch (on/off)
  bool wind_stress;            // Yes/No wind stress switch
  bool ramp;                   // Yes/No ramp function switch
  real ramp_length;            // Length of ramp feature in days
  real ramp_value;             // Current value of the ramp parameter
  real g;                      // Acceleration of free fall: determines units
  const_variable time_step;    // Const/variable time step switch
  real dt;                     // Time step in seconds
  real start_time;             // Starting time of the simulation in days
  real ref_time;               // Reference time in days
  real simul_length;           // Total length of the simulation in days
  int bottom_bc;               // Type of boundary condition at the bottom
  real bottom_fric_coef;       // Friction coefficient for the boundary condition at the bottom
  real bottom_rough_coef;      // Bottom roughness length for the boundary condition at the bottom
  real bottom_l;               // Distance from the computational to the physical boundary at the bottom
  int surface_bc;              // Type of boundary condition at the surface
  real surface_rough_coef;     // Surface roughness length for the boundary condition at the free surface
  real wind_vel_min;           // Limiting (minimum) wind velocity at 10 meters
  real wind_vel_max;           // Limiting (maximum) wind velocity at 10 meters
  real wind_vel[d_space_2d];   // Wind velocity at 10 meters (for the time and space uniform case)
  real surface_stress[d_space_2d]; // Surface stress components (for the time and space uniform case)
  real Rho_air;                // Air density
  real surface_fric_coef[2];   // Surface friction coefficients
  bool steady_state;           // Yes/No steady state/transient simulation switch
  real converg_crit;           // Convergence criterium for steady state simulations
  real unit_conversion_factor; // Unit conversion factor to metric lengths (m/Unit)
  real min_bathym_depth;       // Minimum allowable bathymetric depth (input reset)
  real min_depth;              // Minimum allowable height of water column (run-time reset)
  real ref_lam;                // Reference longitude (in deg) for projection to cartesian coord
  real ref_phi;                // Reference lattitude (in deg) for projection to cartesian coord
  real advect_coef;            // Advection coefficient (default=1, other values work only with the Lax-Friedrichs solver)
  int Riemann_vert_switch;     // Riemann solver (vertical) switch
  int diff_vert_switch;        // Vertical diffusion (eddy viscosity) switch
  int n_diff_vert;             // Number of nonzero vertical diffusion (eddy viscosity) terms
  VAR diff_vert_ind[d_var];    // Index of nonzero vertical diffusion (eddy viscosity) terms
  real diff_vert[d_var];       // Vertical diffusion (eddy viscosity) coefficients
  real diff_vert_max;          // Maximum value of vertical eddy viscosity coefficient (used in turbulence models)
  int diff_horiz_switch;       // Horizontal diffusion (eddy viscosity) switch
  int n_diff_horiz;            // Number of nozero horizontal diffusion (eddy viscosity) terms
  VAR diff_horiz_ind[d_var];   // Index of nozero horizontal diffusion (eddy viscosity) terms
  real diff_horiz[d_var];      // Horizontal diffusion (eddy viscosity) coefficients
  real coriolis_coef;          // Coriolis parameter for the constant Coriolis force case
  int n_tp_freq;               // Number of tidal potential forcing frequencies
  std::vector<tide_info> tp_param;   // Parameters for tidal potential forcing
  int n_bound_tp_freq;         // Number of tidal potential forcing frequencies on open sea boundary 
  std::vector<tide_info> bound_tp_param; // Parameters for tidal pot. forcing on open sea boundaries
  bool elev_rec_stat;          // Output for elevation recording stations switch
  real elev_rec_stat_start;    // Starting time for output for elevation recording stations
  real elev_rec_stat_end;      // End time for output for elevation recording stations
  int elev_rec_stat_freq;      // Output frequency (# of time steps) for elevation recording stations
  int elev_rec_stat_numbr;     // Number of elevation recording stations
  bool vel_rec_stat;           // Output for velocity recording stations switch
  real vel_rec_stat_start;     // Starting time for output for velocity recording stations
  real vel_rec_stat_end;       // End time for output for velocity recording stations
  int vel_rec_stat_freq;       // Output frequency (# of time steps) for velocity recording stations
  int vel_rec_stat_numbr;      // Number of velocity recording stations
  bool elev_global;            // Output for global elevation switch
  real elev_global_start;      // Starting time for output for global elevation
  real elev_global_end;        // End time for output for global elevation
  int elev_global_freq;        // Output frequency (# of time steps) for global elevation
  bool vel_global;             // Output for global velocity switch
  real vel_global_start;       // Starting time for output for global velocity
  real vel_global_end;         // End time for output for global velocity
  int vel_global_freq;         // Output frequency (# of time steps) for global velocity
  bool Rho_dep_pres;           // Density dependence on pressure in the equation of state switch
  real Rho_0;                  // Reference density (at T=0, S=0, z=surface)
#ifdef BAROCLINIC_KEY
  real T_0;                    // Reference temperature (in degrees Celcius)
  real S_0;                    // Reference salinity (in ppt)
#endif

  // Default constructor
  PROBLEM_DESCRIPTION()
  {
    pi=3.141592653589793;
    // advect_coef!=1. works only with the Lax-Friedrichs solver
    advect_coef=1.;
    for (int i=0; i<d_var; diff_horiz[i]=diff_vert[i]=0., i++);
  }

  // Default destructor
  ~PROBLEM_DESCRIPTION(){};

  // Compute cartesian coordinates from spherical
  // Paramters: xi - spherical point coordinates; x - cartesian point coordinates
  void SphericalToCartesian(const real xi[d_space_2d], real x[d_space_2d])
  {
    x[0]=(xi[0]-ref_lam)*Earth_radius*std::cos(ref_phi);
    x[1]=xi[1]*Earth_radius;
    //      coord[0]=(coord_spher[0]-lam_0)*R*std::cos(coord_spher[1]);
    //      coord[1]=(coord_spher[1]-phi_0)*R;
  }

  // Compute spherical coordinates from cartesian
  // Paramters: x - cartesian point coordinates; xi - spherical point coordinates
  void CartesianToSpherical(const real x[d_space_2d], real xi[d_space_2d])
  {
    xi[0]=ref_lam+x[0]/(Earth_radius*std::cos(ref_phi));
    xi[1]=x[1]/Earth_radius;
    //      coord_spher[1]=phi_0+coord[1]/R;
    //      coord_spher[0]=lam_0+coord[0]/(R*std::cos(coord_spher[1]));
  }
};

#if TURBULENCE_KEY==3  // two equation turbulence models
// Storage for the GLS model (Warner et al., 2005)
class GLS_Data
{
public:
  real p, m, n;                // model exponents 
  real kappa;                  // von Karman constant
  real sigma_k, sigma_psi;     // diffusion scaling for k- and psi-equations
  real c1, c2, c3_st, c3_unst; // coefficients for the right hand side of the psi-equation
  real k_min, psi_min;         // Minimum values of k and psi
  real c_mu0, c_mu0_3;         // Model constant
  real C;                      // Model dependent coefficient (=c_mu0^(-p/n))
  real A1, A2, A3, B1, B2;     // Stability function coefficients
  real Gh_min, Gh_max, Gh_c;   // Restrictions on buoyancy parameter (=-N^2l^2/2/k)

  // Constructor
  GLS_Data() {}

  // Model initialization
  // Parameters: model - model choice parameter
  void Init(int model);

  // Stability function computation
  // Parameters: Gh_u - buoyancy parameter (=-N^2l^2/2/k); 
  // S_M - Stability function for the vertical eddy viscosity; S_H - Stability function for the diffusivity
  void Stab(real Gh_u, real& S_M, real& S_H)
  {
    real Gh = Gh_u>Gh_min ? Gh_u : Gh_min;
    Gh = Gh<Gh_c ? Gh : Gh-(Gh-Gh_c)*(Gh-Gh_c)/(Gh+Gh_max-2.*Gh_c);
    S_H=B1/(1.+B2*Gh);
    S_M=(A1+A2*S_H*Gh)/(1.+A3*Gh);
  }

  // Neumann boundary condition for psi variable
  // Parameters: k - turbulent kinetic energy, l - mixing length scale
  // Return value: z-derivative of psi
  real Psi_BC(real k, real l)
  {
    return C*n*std::pow(k, real(m))*kappa*std::pow(kappa*l, real(n-1.));
  }
};
#endif

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
//                                  PDE functions
//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------

// Equation of state for density
// Parameters: coord - point coordinates, var - values of unknowns at the point
// Return value: density
long_real density(const real coord[d_space], const real var[d_var]);

// Advection term computation (for horizontal faces)
// Parameters: var - values of unknowns at the point; advec - result
void advection_horiz(const real var[d_var], long_real advec[d_var][d_space]);

// Advection term computation
// Parameters: var - values of unknowns at the point; advec - result
void advection(const real var[d_var], long_real advec[d_var][d_space]);

// Transport term computation
// Parameters: var - values of unknowns at the point; advec - result
void transport(const real var[d_var], long_real advec[d_var][d_space]);

// Vertical eddy viscosity/diffusivity coefficients computation (constant coefficients)
// Parameters: param - parameters; diff - result
void diffusion_vert_coef_const(const real param[], real diff[d_var][d_space]);

// Vertical eddy viscosity/diffusivity coefficients computation (Davies ~ U*H)
// Parameters: param - parameters; diff - result
void diffusion_vert_coef_uh(const real param[], real diff[d_var][d_space]);

// Vertical eddy viscosity/diffusivity coefficients computation (Davies ~ U*U)
// Parameters: param - parameters; diff - result
void diffusion_vert_coef_uu(const real param[], real diff[d_var][d_space]);

#if TURBULENCE_KEY>1
// Vertical eddy viscosity/diffusivity coef. computation (Xing-Davies k-model)
// Parameters: param - parameters; diff - result
void diffusion_vert_coef_k(const real param[], real diff[d_var][d_space], real *rhs=0);

// Right hand side computation for Xing-Davies k turbulence model
// Parameters: param - parameters; diff - result
// void rhs_k(const real param[], real rhs[d_var]);
#endif


#if TURBULENCE_KEY==3
// Vertical eddy viscosity/diffusivity coef. computation for the GLS model (Warner et al., 2005)
// Parameters: param - parameters; diff - result
void diffusion_vert_coef_gls(const real param[], real diff[d_var][d_space], real *rhs=0);


// Right hand side computation for the GLS model (Warner et al., 2005)
// Parameters: param - parameters; diff - result
// void rhs_gls(const real param[], real rhs[d_var]);
#endif

// Horizontal eddy viscosity/diffusivity coefficients computation (constant coefficients)
// Parameters: param - parameters; diff - result
void diffusion_horiz_coef(const real param[], real diff[d_var][d_space]);




//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
//                                   Riemann solvers
//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------

// Parameters: normal - normal to the face;
//             var_l, var_r - values of the unknowns on both sides of the face; flux - result
// Normal advective flux solver for Riemann problem on vertical faces
void advection_vert_flux(const long_real normal[d_space], const real var_l[d_var],
                         const real var_r[d_var], long_real flux[d_var]);

// HLL solver for Riemann problem on vertical faces
void HLL(const long_real normal[d_space], const real var_l[d_var], const real var_r[d_var],
         long_real flux[d_var]);

// HLLC solver for Riemann problem on vertical faces
void HLLC(const long_real normal[d_space], const real var_l[d_var], const real var_r[d_var],
          long_real flux[d_var]);

// Roe's solver for Riemann problem on vertical faces
void Roe(const long_real normal[d_space], const real var_l[d_var], const real var_r[d_var],
         long_real flux[d_var]);

// Roe's solver for Riemann problem on vertical faces with an entropy fix
void RoeEntropyFix(const long_real normal[d_space], const real var_l[d_var],
                   const real var_r[d_var], long_real flux[d_var]);

// Lax-Friedrichs solver for Riemann problem on vertical faces
void LaxFriedrichs(const long_real normal[d_space], const real var_l[d_var],
                   const real var_r[d_var], long_real flux[d_var]);

// FORCE solver for Riemann problem on vertical faces
void FORCE(const long_real normal[d_space], const real var_l[d_var], const real var_r[d_var],
           long_real flux[d_var]);
#endif
