#include <cmath>
#include "pde.H"
#include "test.H"

extern PROBLEM_DESCRIPTION problem_info;

const real freq=0.01, elev_coef=0.01, amp=0.1, t_coef=1.;

// True solution for the test problem
void solution_analytic(real t, const real coord[d_space], real func[d_var])
{
  real c0=std::cos(freq*(coord[0]+t_coef*t)), c1=std::cos(freq*(coord[1]+t_coef*t)),
    s0=std::sin(freq*(coord[0]+t_coef*t)), s1=std::sin(freq*(coord[1]+t_coef*t)), 
    s2=std::sin(freq*(coord[2]+t_coef*t));

  func[U]=amp*(s0+s2);
  func[V]=amp*(s1+s2);
  func[W]=-amp*freq*coord[2]*(c0+c1);
}

// Gradient of the true solution
void solution_analytic_grad(real t, const real coord[d_space], real grad[d_var][d_space])
{
  real coef=amp*freq, c0=std::cos(freq*(coord[0]+t_coef*t)), 
    c1=std::cos(freq*(coord[1]+t_coef*t)), c2=std::cos(freq*(coord[2]+t_coef*t)),
    s0=std::sin(freq*(coord[0]+t_coef*t)), s1=std::sin(freq*(coord[1]+t_coef*t));

  grad[U][0]=coef*c0;
  grad[U][1]=0.;
  grad[U][2]=coef*c2;
  grad[V][0]=0.;
  grad[V][1]=coef*c1;
  grad[V][2]=coef*c2;
  grad[W][0]=coef*freq*coord[2]*s0;
  grad[W][1]=coef*freq*coord[2]*s1;
  grad[W][2]=-coef*(c0+c1);
}

// Laplacian of the true solution
void solution_analytic_laplacian(real t, const real coord[d_space], real lapl[d_var][d_space])
{
  real coef=amp*freq*freq, s0=std::sin(freq*(coord[0]+t_coef*t)), 
    s1=std::sin(freq*(coord[1]+t_coef*t)), s2=std::sin(freq*(coord[2]+t_coef*t));

  lapl[U][0]=-coef*s0;
  lapl[U][1]=0.;
  lapl[U][2]=-coef*s2;
  lapl[V][0]=0.;
  lapl[V][1]=-coef*s1;
  lapl[V][2]=-coef*s2;
}

// Time derivative of the true solution for the test problem
void solution_analytic_dt(real t, const real coord[d_space], real t_deriv[d_var])
{
  real coef=t_coef*amp*freq;

  t_deriv[U]=coef*(std::cos(freq*(coord[0]+t_coef*t))+std::cos(freq*(coord[2]+t_coef*t)));
  t_deriv[V]=coef*(std::cos(freq*(coord[1]+t_coef*t))+std::cos(freq*(coord[2]+t_coef*t)));
}

// Analytic solution for the elevation
void solution_analytic_elevation(real t, const real coord[2], real result[])
{
  result[0]=elev_coef*(std::sin(freq*(coord[0]+t_coef*t))+std::sin(freq*(coord[1]+t_coef*t)));
}

// Analytic solution for the elevation gradient
void solution_analytic_elevation_grad(real t, const real coord[2], real elev_grad[2])
{
  elev_grad[0]=elev_coef*freq*std::cos(freq*(coord[0]+t_coef*t));
  elev_grad[1]=elev_coef*freq*std::cos(freq*(coord[1]+t_coef*t));
}

// Time derivative of the true solution for the elevation
void solution_analytic_elevation_dt(real t, const real coord[d_space], real result[])
{
  result[0]=t_coef*elev_coef*freq*(std::cos(freq*(coord[0]+t_coef*t))+std::cos(freq*(coord[1]+t_coef*t)));
}

#ifdef BAROCLINIC_KEY
// Equation of state for density
long_real density(const real coord[d_space], const real var[d_var])
{
  return amp*coord[0]/problem_info.g;
}

// Forcing function accounting for density terms
void density_forcing(real t, real xi, const real xi_grad[d_space_2d], 
                     const real coord[d_space], real result[])
{
  real var[d_var], coord_top[d_space]={coord[0], coord[1], xi}, ro=density(coord_top, var);
  real coef=0.2*problem_info.pi;

  result[0]=amp*(xi-coord[2])+problem_info.g*ro*xi_grad[0];
  result[1]=problem_info.g*ro*xi_grad[1];
}
#endif
