#include <iostream>
#include <fstream>
#include <cstring>
#include <string>
#include <sstream>
#include <cassert>
#include <exception>
#include "recording_stations.H"
#include "pde.H"
#include "basis.H"
#include "utility.H"
#include "mesh_container.H"

extern BASIS_2D *basis_2d[max_order+1];   // 2D basis objects
extern BASIS_REF *basis_ref[max_order+1]; // 3D basis objects for parallel prisms
extern DOF var_stage[max_stages][d_var];  // Global storage for degrees of freedom for Runge-Kutta stages
extern DOF var_rhs[d_var];                // Global storage for degrees of freedom for the right hand side
extern DOF var_flux[d_var][d_space];      // Global storage for degrees of freedom for fluxes
extern DOF var[d_var];                    // Global storage for degrees of freedom for unknowns
extern PROBLEM_DESCRIPTION problem_info;  // Storage for physical parameters
extern MESH mesh;                         // Mesh object
#if TURBULENCE_KEY==3  // two equation turbulence models
extern GLS_Data gls_data;                 // Data for two equation turbulence models
#endif


// Input a data field from an input file
// Parameters: buffer - input buffer; format - data marker to look for; data - input data, lvalue
template<typename T>
void InputDataField(const std::string &buffer, const std::string &format, T &data)
{
  std::string::size_type pos=buffer.find(format);
  if (pos!=std::string::npos)
  {
    T tmp;
    std::istringstream strin(buffer.substr(pos+format.length()));

    strin >> tmp;
    if (!strin.fail())
      data=tmp;
  }
}


// Read configuration file
// Parameters: config_file - name of the main configuration file; t - time, lvalue
void InputConfig(const char* config_file, long_real& t)
{
  const int buffer_size=300;
  int inp;
  real max_depth=0., tmp;
  char buffer[buffer_size], grid_coord[buffer_size], simulation_param[buffer_size], 
    grid_connect[buffer_size];
  std::ifstream config(config_file);
  std::string strng;

  if (!config)
  {
    std::cerr << "File utility.cpp: The configuration file could not be found!\n";
    std::exit(1);
  }

  try
  {
    // Read header
    config.getline(buffer, buffer_size);
    config.ignore(buffer_size, '\n');
    if (std::strcmp(buffer, "Configuration file for UTBEST_3D"))
    {
      std::cerr << "File utility.cpp: Error in configuration file header!\n";
      std::exit(1);
    }

    // Read names of grid input files
    config.get(grid_coord, buffer_size, ' ');
    config.ignore(buffer_size, '\n');
    config.get(simulation_param, buffer_size, ' ');
    config.ignore(buffer_size, '\n');
    config.get(grid_connect, buffer_size, ' ');
    config.ignore(buffer_size, '\n');
    config.get(problem_info.output_file, sizeof(problem_info.output_file), ' ');
    config.ignore(buffer_size, '\n');
    std::strcpy(problem_info.tecplot_out, problem_info.output_file);
    std::strcat(problem_info.tecplot_out, ".out");

    // Input vertical diffusion (eddy viscosity) model switch and coefficients 
    problem_info.n_diff_vert=0;
    config >> problem_info.diff_vert_switch;
    std::getline(config, strng);
    strng.erase(strng.find('!'));
    // Input maximum value of the vertical eddy viscosity coefficient
    problem_info.diff_vert_max=1.;
    InputDataField(strng, std::string("Max="), problem_info.diff_vert_max);
    if (problem_info.diff_vert_max<0)
    {
      std::cerr << "File utility.cpp: Inadmissible value of the maximum eddy viscosity coefficient!\n";
      std::exit(1);
    }
    switch(problem_info.diff_vert_switch)
    {
    case 0: 
      break;
    case 2:  // Algebraic turbulence models
    case 3:
#ifndef DEPTH_AVERAGED_VELOCITY_KEY
      std::cerr << "File utility.cpp: Attempting to use algebraic eddy visosity model "
                << "with disabled depth averaged velocity computation!\n";
      std::exit(1);
#endif
   case 1:
      {
        for (int i=0; i<d_var_3d_dt; i++)
        {
          std::string format=" ";
          format+=VAR_NAME[VAR_3D_DT_INDEX[i]];
          format+="=";
          InputDataField(strng, format, problem_info.diff_vert[VAR_3D_DT_INDEX[i]]);
          if (problem_info.diff_vert[VAR_3D_DT_INDEX[i]]!=0.)
            problem_info.diff_vert_ind[problem_info.n_diff_vert++]=VAR_3D_DT_INDEX[i];
        }
      }
      break;
#if TURBULENCE_KEY>1  // Higher order turbulence models
#if TURBULENCE_KEY==3  // two equation turbulence models
    case 5:  // Two equation turbulence models
    case 6:
    case 7:
    case 8: gls_data.Init(problem_info.diff_vert_switch);
#endif
    case 4:  // One equation turbulence models
      problem_info.n_diff_vert=d_var_3d_dt;
      for (int i=0; i<d_var_3d_dt; i++)
      {
        problem_info.diff_vert[VAR_3D_DT_INDEX[i]]=1.;
        problem_info.diff_vert_ind[i]=VAR_3D_DT_INDEX[i];
      }
      break;
#endif
    default: std::cerr << "File utility.cpp: Error in vertical eddy viscosity switch parameter!\n";
      std::exit(1);
    }

    // Input bottom boundary conditions
    {
      problem_info.bottom_fric_coef=0.;
      problem_info.bottom_rough_coef=0.005;
      config >> problem_info.bottom_bc;
      std::getline(config, strng);
      strng.erase(strng.find('!'));
      // Input bottom friction coefficient
      InputDataField(strng, std::string("Fric="), problem_info.bottom_fric_coef);
      // Input bottom roughness coefficient
        InputDataField(strng, std::string("Rough="), problem_info.bottom_rough_coef);
      problem_info.bottom_l=0.;
#if TURBULENCE_KEY==3  // Higher order turbulence models
      if (problem_info.diff_vert_switch>4 && problem_info.bottom_fric_coef!=0.)
        problem_info.bottom_l=problem_info.bottom_rough_coef*
          std::exp(gls_data.kappa/std::sqrt(problem_info.bottom_fric_coef));
#endif
    }

    // Input surface boundary conditions
    {
      problem_info.surface_fric_coef[0]=0.61;
      problem_info.surface_fric_coef[1]=0.063;
      problem_info.surface_rough_coef=0.0;
      problem_info.wind_vel_min=6.;
      problem_info.wind_vel_max=50.;
      problem_info.wind_vel[0]=problem_info.wind_vel[1]=0.;
      problem_info.surface_stress[0]=problem_info.surface_stress[1]=0.;
      config >> problem_info.surface_bc;
      std::getline(config, strng);
      strng.erase(strng.find('!'));
      switch (problem_info.surface_bc)
      {
      case 0: break;  // No wind
      case 1:         // Input wind velocity
        InputDataField(strng, std::string("U="), problem_info.wind_vel[0]);
        InputDataField(strng, std::string("V="), problem_info.wind_vel[1]);
        // Compute surface stress from velocity
        {
          real U_abs=std::sqrt(problem_info.wind_vel[0]*problem_info.wind_vel[0]+
                               problem_info.wind_vel[1]*problem_info.wind_vel[1]);
          U_abs = U_abs<problem_info.wind_vel_min ? problem_info.wind_vel_min : 
            U_abs>problem_info.wind_vel_max ? problem_info.wind_vel_max : U_abs;
          real fric_coef=0.001*problem_info.Rho_air/problem_info.Rho_0*
            (problem_info.surface_fric_coef[0]+U_abs*problem_info.surface_fric_coef[1]);
          for (int i=0; i<d_space_2d; i++)
            problem_info.surface_stress[i]=fric_coef*U_abs*problem_info.wind_vel[i];
        }
        break;
      case 2:         // Input surface stress
        InputDataField(strng, std::string("tau_x="), problem_info.surface_stress[0]);
        InputDataField(strng, std::string("tau_y="), problem_info.surface_stress[1]);
        break;
      default: std::cerr << "File utility.cpp: unknown surface boundary condition!\n";
      }
      // Input surface roughness coefficient
      InputDataField(strng, std::string("Rough="), problem_info.surface_rough_coef);
    }

    // Input baroclinic data
    config >> inp;
    if (inp!=0 && inp!=1)
    {
      std::cerr << "File utility.cpp: Inadmissible density dependence on pressure parameter!\n";
      std::exit(1);
    }
    problem_info.Rho_dep_pres = inp==0 ? false : true;
    problem_info.Rho_air=0.;
    problem_info.Rho_0=0.;
    std::getline(config, strng);
    strng.erase(strng.find('!'));
    // Input, if available, reference values of water and air density
    InputDataField(strng, std::string("Rho_0="), problem_info.Rho_0);
    InputDataField(strng, std::string("Rho_air="), problem_info.Rho_air);
#ifdef BAROCLINIC_KEY // Input, if available, reference values of temperature and salinity
    problem_info.T_0=20.;
    problem_info.S_0=35.;
    InputDataField(strng, std::string("T_0="), problem_info.T_0);
    InputDataField(strng, std::string("S_0="), problem_info.S_0);
#endif

    // Input number of layers and mesh grading at the bottom
    problem_info.mesh_grading_bot=0.;
    config >> problem_info.n_layers;
    if (problem_info.n_layers<1)
    {
      std::cerr << "File utility.cpp: Inadmissible number of vertical layers!\n";
      std::exit(1);
    }
    std::getline(config, strng);
    strng.erase(strng.find('!'));
    InputDataField(strng, std::string("Grad="), problem_info.mesh_grading_bot);
    if (problem_info.mesh_grading_bot<0)
    {
      std::cerr << "File utility.cpp: Error in mesh grading at the bottom parameter!\n";
      std::exit(1);
    }

    // Input frequency of mesh update
    config >> problem_info.mesh_update;
    if (problem_info.mesh_update<-1)
    {
      std::cerr << "File utility.cpp: Inadmissible value of free surface update frequency!\n";
      std::exit(1);
    }
    config.ignore(buffer_size, '\n');

    // Input adaptivity info
    config >> problem_info.max_refinement;
    if (problem_info.max_refinement<0)
    {
      std::cerr << "File utility.cpp: Inadmissible number of mesh refinement levels!\n";
      std::exit(1);
    }
    std::getline(config, strng);
    strng.erase(strng.find('!'));
    problem_info.refinement_var=W;
#ifdef ADAPTIVITY_H_2D_KEY
    for (int i=0; i<d_var_2d_dt; i++)
    {
      std::string format=" ";
      format+=VAR_NAME[VAR_2D_DT_INDEX[i]];
      format+=" ";
      if (strng.find(format)!=std::string::npos)
        {
          problem_info.refinement_var=VAR_2D_DT_INDEX[i];
          break;
        }
    }
    if (problem_info.refinement_var==W)
    {
      std::cerr << "File utility.cpp: Variable name for shock detector not found!\n"
                << "Check VAR_NAME and adaptivity keys in general.H!\n";
      std::exit(1);
    }
#endif
#ifdef ADAPTIVITY_H_3D_KEY
    for (int i=0; i<d_var_3d_dt; i++)
    {
      std::string format=" ";
      format+=VAR_NAME[VAR_3D_DT_INDEX[i]];
      format+=" ";
      if (strng.find(format)!=std::string::npos)
        {
          problem_info.refinement_var=VAR_3D_DT_INDEX[i];
          break;
        }
    }
    if (problem_info.refinement_var==W)
    {
      std::cerr << "File utility.cpp: Variable name for shock detector not found!\n"
                << "Check VAR_NAME and adaptivity keys in general.H!\n";
      std::exit(1);
    }
#endif
    InputDataField(strng, std::string("t_0="), problem_info.refinement_start);
    problem_info.refinement_start*=86400.;
    InputDataField(strng, std::string("shock="), problem_info.detector_sensitivity);

    // Input Riemann solver choice parameter
    config >> problem_info.Riemann_vert_switch;
    config.ignore(buffer_size, '\n');

    // Input Runge-Kutta method choice parameter
    config >> problem_info.runge_kutta_method;
    switch(problem_info.runge_kutta_method)
    {
    case 1: case 2: case 3: break;
    default:
      std::cerr << "File utility.cpp: Error in Runge-Kutta method choice parameter!\n";
      std::exit(1);
    }
    config.ignore(buffer_size, '\n');

    // Input order of approximation
    config >> problem_info.def_order;
    if ((problem_info.def_order<0) || (problem_info.def_order>max_order))
    {
      std::cerr << "File utility.cpp: Inadmissable order of approximating space!\n";
      std::exit(1);
    }
    std::getline(config, strng);
    strng.erase(strng.find('!'));
    for (int i=0; i<d_var; i++)
    {
      std::string format=" ";
      format+=VAR_NAME[i];
      format+="=";
      problem_info.var_order[i]=problem_info.def_order;
      InputDataField(strng, format, problem_info.var_order[i]);
      if (problem_info.var_order[i]!=problem_info.def_order)
        if (problem_info.var_order[i]<0 || problem_info.var_order[i]>max_order)
        {
          std::cerr << "File utility.cpp: Inadmissable order of approximating space!\n";
          std::exit(1);
        }
    }
    config.close();
  }
  catch(std::exception& ex)
  {
    std::cerr << "File utility.cpp: I/O error in configuration file" << ex.what() << '\n';
    std::exit(1);
  }

  // Input of simulation parameters (format compatible with UTBEST)
  std::ifstream param(simulation_param);

  if (!param)
  {
    std::cerr << "File utility.cpp: The simulation parameters file " << simulation_param << 
      " could not be found!\n";
    std::exit(1);
  }

  try
  {
    param >> std::ws;
    param.get(problem_info.run_description, sizeof(problem_info.run_description), ' ');
    param.ignore(buffer_size, '\n');
    param >> std::ws;
    param.get(problem_info.run_id, sizeof(problem_info.run_id), ' ');
    param.ignore(buffer_size, '\n');
    param >> problem_info.hot_input;
    switch (problem_info.hot_input)
    {
    case 0: break;
    case 1: param >> std::ws;
      param.get(problem_info.hot_input_file, buffer_size, ' ');
      break;
    default: std::cerr << "File utility.cpp: Cold/Hot start parameter out of range!\n";
      std::exit(1);
    }
    param.ignore(buffer_size, '\n');
    param >> inp;
    param.ignore(buffer_size, '\n');
    switch (inp)
    {
    case 1: problem_info.coord_system=PROBLEM_DESCRIPTION::CARTESIAN;
      break;
    case 2: problem_info.coord_system=PROBLEM_DESCRIPTION::SPHERICAL;
      break;
    default: std::cerr << "File utility.cpp: Coordinate system parameter out of range!\n";
      std::exit(1);
    }
    param.ignore(buffer_size, '\n');
    param.ignore(buffer_size, '\n');
    param >> inp;
    param.ignore(buffer_size, '\n');
    switch (inp)
    {
    case 0: problem_info.coriolis_par=PROBLEM_DESCRIPTION::CONSTANT;
      break;
    case 1: problem_info.coriolis_par=PROBLEM_DESCRIPTION::VARIABLE;
      break;
    default: std::cerr << "File utility.cpp: Coriolis const/variable switch parameter out of range!\n";
      std::exit(1);
    }
    param >> inp;
    param.ignore(buffer_size, '\n');
    switch (inp)
    {
    case 0: problem_info.tidal_poten=false;
      break;
    case 1: problem_info.tidal_poten=true;
      break;
    default: std::cerr << "File utility.cpp: Tidal potential const/variable switch parameter out of range!\n";
      std::exit(1);
    }
    param >> inp;
    param.ignore(buffer_size, '\n');
    switch (inp)
    {
    case 0: problem_info.wind_stress=false;
      break;
    case 1: problem_info.wind_stress=true;
      break;
    default: std::cerr << "File utility.cpp: Wind stress on/off switch parameter out of range!\n";
      std::exit(1);
    }
    param >> inp;
    param.ignore(buffer_size, '\n');
    switch (inp)
    {
    case 0: problem_info.ramp=false; problem_info.ramp_value=1.; break;
    case 1: problem_info.ramp=true; problem_info.ramp_value=0.; break;
    default: std::cerr << "File utility.cpp: Ramp on/off switch parameter out of range!\n";
      std::exit(1);
    }
    // Input free fall acceleration (determines units)
    param >> problem_info.g;
    param.ignore(buffer_size, '\n');
    problem_info.unit_conversion_factor=problem_info.g/9.81;
#if TURBULENCE_KEY==3 // Units conversion for turbulence unknowns
    if (problem_info.diff_vert_switch>4)
     {
       gls_data.k_min*=std::pow(problem_info.unit_conversion_factor, 2*gls_data.m);
       gls_data.psi_min*=std::pow(problem_info.unit_conversion_factor, 2*gls_data.m+gls_data.n);
     }
#endif
    if (problem_info.Rho_air==0.)
      problem_info.Rho_air=1.204/problem_info.unit_conversion_factor
        /problem_info.unit_conversion_factor/problem_info.unit_conversion_factor;
    if (problem_info.Rho_0==0.)
      problem_info.Rho_0=999.83/problem_info.unit_conversion_factor
        /problem_info.unit_conversion_factor/problem_info.unit_conversion_factor;
// #ifdef BAROCLINIC_KEY // Input, if available, reference values of density, temperature, and salinity
//     {
//       problem_info.T_0=20.;
//       problem_info.S_0=35.;
//       param.getline(buffer, buffer_size);
//       std::string strng(buffer);
//       std::istringstream strin(strng);
//       strin >> tmp;
//       if (!strin.fail())
//         problem_info.Rho_0=tmp;
//       strin >> tmp;
//       if (!strin.fail())
//         problem_info.T_0=tmp;
//       strin >> tmp;
//       if (!strin.fail())
//         problem_info.S_0=tmp;
//     }
// #else
//     param.ignore(buffer_size, '\n');
// #endif
    // Compute surface stress from wind velocity
    if (problem_info.surface_bc==1)
    {
      real U_abs=std::sqrt(problem_info.wind_vel[0]*problem_info.wind_vel[0]+
                           problem_info.wind_vel[1]*problem_info.wind_vel[1]);
      U_abs = U_abs<problem_info.wind_vel_min ? problem_info.wind_vel_min : 
        U_abs>problem_info.wind_vel_max ? problem_info.wind_vel_max : U_abs;
      real fric_coef=0.001*problem_info.Rho_air/problem_info.Rho_0*
        (problem_info.surface_fric_coef[0]+U_abs*problem_info.surface_fric_coef[1]);
      for (int i=0; i<d_space_2d; i++)
        problem_info.surface_stress[i]=fric_coef*U_abs*problem_info.wind_vel[i];
    }

    // Convert Earth radius to correct units
    problem_info.Earth_radius=6378206.4*problem_info.unit_conversion_factor;
    param >> inp;
    param.ignore(buffer_size, '\n');
    for (int i=0; i<inp; i++)
      param.ignore(buffer_size, '\n');
    param >> inp;
    param.ignore(buffer_size, '\n');
    switch (inp)
    {
    case 0: problem_info.time_step=PROBLEM_DESCRIPTION::CONSTANT;
      break;
    case 1: problem_info.time_step=PROBLEM_DESCRIPTION::VARIABLE;
      break;
    default: std::cerr << "File utility.cpp: Time step Const/Variable switch parameter out of range!\n";
      std::exit(1);
    }
    param >> problem_info.dt;
    param.ignore(buffer_size, '\n');
    param >> problem_info.start_time;
    param.ignore(buffer_size, '\n');
    param >> problem_info.ref_time;
    param.ignore(buffer_size, '\n');
    param >> problem_info.simul_length;
    param.ignore(buffer_size, '\n');
    param.ignore(buffer_size, '\n');
    param.ignore(buffer_size, '\n');
    param >> inp;
    switch (inp)
    {
    case 0: problem_info.steady_state=false;
      break;
    case 1: problem_info.steady_state=true;
      break;
    default: std::cerr << "File utility.cpp: Transient/steady switch parameter out of range!\n";
      std::exit(1);
    }
    param.ignore(buffer_size, '\n');
    param >> problem_info.converg_crit;
    param.ignore(buffer_size, '\n');
    param >> problem_info.ramp_length;
    param.ignore(buffer_size, '\n');
    param >> problem_info.min_depth;
    {
      problem_info.min_bathym_depth=problem_info.min_depth;
      param.getline(buffer, buffer_size);
      std::string strng(buffer);
      std::istringstream strin(strng);
      strin >> tmp;
      if (!strin.fail())
        problem_info.min_bathym_depth=tmp;
    }
    param >> problem_info.ref_lam >> problem_info.ref_phi;
    // Convert lam and phi from degrees to radiants
    problem_info.ref_lam*=problem_info.pi/180.;
    problem_info.ref_phi*=problem_info.pi/180.;
    param.ignore(buffer_size, '\n');
    param.ignore(buffer_size, '\n');
    param.ignore(buffer_size, '\n');
    param >> problem_info.diff_horiz_switch;
    if (problem_info.diff_horiz_switch<0 || problem_info.diff_horiz_switch>1)
    {
      std::cerr << "File utility.cpp: Error in horizontal eddy viscosity switch parameter!\n";
      std::exit(1);
    }
    param.ignore(buffer_size, '\n');
    param >> problem_info.diff_horiz[U];
    problem_info.diff_horiz[V]=problem_info.diff_horiz[U];
#ifdef BAROCLINIC_KEY // Input, if available, diffusion coefficients for temperature and salinity
    {
      param.getline(buffer, buffer_size);
      std::string strng(buffer);
      std::istringstream strin(strng);
      strin >> tmp;
      if (!strin.fail())
        problem_info.diff_horiz[T]=tmp;
      strin >> tmp;
      if (!strin.fail())
        problem_info.diff_horiz[S]=tmp;
    }
#else
    param.ignore(buffer_size, '\n');
#endif
    problem_info.n_diff_horiz=0;
    if (problem_info.diff_horiz_switch!=0)
      for (int i=0; i<d_var_3d_dt; i++)
        if (problem_info.diff_horiz[VAR_3D_DT_INDEX[i]]!=0.)
          problem_info.diff_horiz_ind[problem_info.n_diff_horiz++]=VAR_3D_DT_INDEX[i];
    param >> problem_info.coriolis_coef;
    param.ignore(buffer_size, '\n');
    param >> problem_info.n_tp_freq;
    param.ignore(buffer_size, '\n');
    if ((problem_info.n_tp_freq>0 && !problem_info.tidal_poten) ||
      (!problem_info.n_tp_freq && problem_info.tidal_poten))
    {
      std::cerr << "File utility.cpp: Contradicting information regarding tidal potential forcing!\n";
      std::exit(1);
    }
    if (problem_info.n_tp_freq>0)
    {
      problem_info.tp_param.resize(problem_info.n_tp_freq);
      for (int i=0; i<problem_info.n_tp_freq;i++)
      {
        param >> std::ws;
        param.get(problem_info.tp_param[i].tp_tag, 5, '\n');
        param.ignore(buffer_size, '\n');
        param >> problem_info.tp_param[i].tp_coef 
          >> problem_info.tp_param[i].amig
          >> problem_info.tp_param[i].etrf
          >> problem_info.tp_param[i].ff
          >> problem_info.tp_param[i].phase;
        problem_info.tp_param[i].phase*=problem_info.pi/180.;
        param.ignore(buffer_size, '\n');
      }
    }
    param >> problem_info.n_bound_tp_freq;
    param.ignore(buffer_size, '\n');
    if (problem_info.n_bound_tp_freq>0)
    {
      problem_info.bound_tp_param.resize(problem_info.n_bound_tp_freq);
      for (int i=0; i<problem_info.n_bound_tp_freq;i++)
      {
        param >> std::ws;
        param.get(problem_info.bound_tp_param[i].tp_tag, 5, '\n');
        param.ignore(buffer_size, '\n');
        param >> problem_info.bound_tp_param[i].amig 
          >> problem_info.bound_tp_param[i].ff 
          >> problem_info.bound_tp_param[i].phase;
        problem_info.bound_tp_param[i].phase*=problem_info.pi/180.;
        param.ignore(buffer_size, '\n');
      }
    }

    // Read coarse mesh
    if (problem_info.hot_input==0)
    {
      mesh.Input2DMesh(grid_coord, grid_connect, max_depth);
      mesh.Create3DMesh(max_depth);
      t=problem_info.start_time;
    }
    else
      mesh.HotInput(t);

    // Input elevation recording stations
    param >> inp >> problem_info.elev_rec_stat_start >> problem_info.elev_rec_stat_end >> 
      problem_info.elev_rec_stat_freq;
    param.ignore(buffer_size, '\n');
    switch (inp)
    {
    case 0: problem_info.elev_rec_stat=false;
      break;
    case 1: problem_info.elev_rec_stat=true;
      break;
    default: std::cerr << "File utility.cpp: Elevation recording station switch parameter out of range!\n";
      std::exit(1);
    }
    if(problem_info.elev_rec_stat_end-problem_info.elev_rec_stat_start<0.)
    {
      std::cerr << "File utility.cpp: End of elevation recording station output comes before its begin!\n";
      std::exit(1);
    }
    param >> problem_info.elev_rec_stat_numbr;
    param.ignore(buffer_size, '\n');
    if (problem_info.elev_rec_stat_numbr<0)
    {
      std::cerr << "File utility.cpp: Number of elevation recording stations out of range!\n";
      std::exit(1);
    }
    // Initialize elevation recording stations
    if (problem_info.elev_rec_stat && (problem_info.elev_rec_stat_numbr>0))
    {
      int n_spools=(int)((problem_info.elev_rec_stat_end-problem_info.elev_rec_stat_start)*
        86400./(problem_info.elev_rec_stat_freq*problem_info.dt));

      std::strcpy(problem_info.elev_out, problem_info.output_file);
      std::strcat(problem_info.elev_out, "_elev.out");

      // Output header
      if (mesh.IterationCounter()==0)
      {
        std::ofstream elev_out(problem_info.elev_out, std::ofstream::trunc);

        elev_out << "   " << problem_info.run_description
          << "   " << problem_info.run_id 
          << "   " << problem_info.grid_desc << std::endl
          << "   " << n_spools << "   " << problem_info.elev_rec_stat_numbr
          << "   " << problem_info.elev_rec_stat_freq*problem_info.dt
          << "   "<< problem_info.elev_rec_stat_freq << "   " << 1 << std::endl;
        elev_out.close();
      }
      VAR var_ind[]={XI};
      mesh.elev_stations.InitializeStations(problem_info.elev_rec_stat_numbr, 
        problem_info.elev_rec_stat_freq, problem_info.elev_rec_stat_start, 
        problem_info.elev_rec_stat_end, problem_info.elev_out, var_ind);
      for (int i=0; i<problem_info.elev_rec_stat_numbr; i++)
      {
        real coord[d_space]={0.};
        param >> coord[0] >> coord[1];
        param.ignore(buffer_size, '\n');
        if (problem_info.coord_system==PROBLEM_DESCRIPTION::SPHERICAL)
        {
          real coord_tmp[d_space_2d];
          coord[0]*=problem_info.pi/180.;
          coord[1]*=problem_info.pi/180.;
          problem_info.SphericalToCartesian(coord, coord_tmp);
          mesh.elev_stations.AssignCoord(i, coord_tmp);
        }
        else
          mesh.elev_stations.AssignCoord(i, coord);
      }
    }
    // Assign elevation recording stations to elements
    mesh.elev_stations.AssignStations(mesh.triangle_2d_list);

    // Input velocity recording stations
    param >> inp >> problem_info.vel_rec_stat_start >> problem_info.vel_rec_stat_end >> 
      problem_info.vel_rec_stat_freq;
    param.ignore(buffer_size, '\n');
    switch (inp)
    {
    case 0: problem_info.vel_rec_stat=false;
      break;
    case 1: problem_info.vel_rec_stat=true;
      break;
    default: std::cerr << "File utility.cpp: Velocity recording station switch parameter out of range!\n";
      std::exit(1);
    }
    if(problem_info.vel_rec_stat_end-problem_info.vel_rec_stat_start<0.)
    {
      std::cerr << "File utility.cpp: End of velocity recording station output comes before its begin!\n";
      std::exit(1);
    }
    param >> problem_info.vel_rec_stat_numbr;
    param.ignore(buffer_size, '\n');
    if (problem_info.vel_rec_stat_numbr<0)
    {
      std::cerr << "File utility.cpp: Number of velocity recording stations out of range!\n";
      std::exit(1);
    }
    // Initialize velocity recording stations
    if (problem_info.vel_rec_stat && (problem_info.vel_rec_stat_numbr>0))
    {
      int n_spools=(int)((problem_info.vel_rec_stat_end-problem_info.vel_rec_stat_start)*
        86400./(problem_info.vel_rec_stat_freq*problem_info.dt));

#ifndef DEPTH_AVERAGED_VELOCITY_KEY
      std::cerr << "File utility.cpp: Depth averaged velocity computation not supported!\n"
                << "Enable DEPTH_AVERAGED_VELOCITY_KEY in general.H or remove velocity recording stations!\n";
      std::exit(1);
#else

      std::strcpy(problem_info.vel_out, problem_info.output_file);
      std::strcat(problem_info.vel_out, "_vel.out");

      // Output header
      if (mesh.IterationCounter()==0)
      {
        std::ofstream vel_out(problem_info.vel_out, std::ofstream::trunc);

        vel_out << "   " << problem_info.run_description
          << "   " << problem_info.run_id
          << "   " << problem_info.grid_desc << std::endl
          << "   " << n_spools << "   " << problem_info.vel_rec_stat_numbr
          << "   " << problem_info.vel_rec_stat_freq*problem_info.dt
          << "   "<< problem_info.vel_rec_stat_freq << "   " << 2 << std::endl;
        vel_out.close();
      }
      VAR var_ind[]={UH, VH};
      mesh.vel_stations.InitializeStations(problem_info.vel_rec_stat_numbr, 
        problem_info.vel_rec_stat_freq, problem_info.vel_rec_stat_start, 
        problem_info.vel_rec_stat_end, problem_info.vel_out, var_ind);
      for (int i=0; i<problem_info.vel_rec_stat_numbr; i++)
      {
        real coord[d_space]={0.};
        param >> coord[0] >> coord[1];
        param.ignore(buffer_size, '\n');
        if (problem_info.coord_system==PROBLEM_DESCRIPTION::SPHERICAL)
        {
          real coord_tmp[d_space_2d];
          coord[0]*=problem_info.pi/180.;
          coord[1]*=problem_info.pi/180.;
          problem_info.SphericalToCartesian(coord, coord_tmp);
          mesh.vel_stations.AssignCoord(i, coord_tmp);
        }
        else
          mesh.vel_stations.AssignCoord(i, coord);
      }
#endif
    }
    // Assign velocity recording stations to elements
    mesh.vel_stations.AssignStations(mesh.triangle_2d_list);

    param >> inp >> problem_info.elev_global_start >> problem_info.elev_global_end >> 
      problem_info.elev_global_freq;
    param.ignore(buffer_size, '\n');
    switch (inp)
    {
    case 0: problem_info.elev_global=false;
      break;
    case 1: problem_info.elev_global=true;
      break;
    default: std::cerr << "File utility.cpp: Elevation recording station switch parameter out of range!\n";
      std::exit(1);
    }
    if(problem_info.elev_global_end-problem_info.elev_global_start<0.)
    {
      std::cerr << "File utility.cpp: End of elevation recording station output comes before its begin!\n";
      std::exit(1);
    }
    param >> inp >> problem_info.vel_global_start >> problem_info.vel_global_end >> 
      problem_info.vel_global_freq;
    param.ignore(buffer_size, '\n');
    switch (inp)
    {
    case 0: problem_info.vel_global=false;
      break;
    case 1: problem_info.vel_global=true;
      break;
    default: std::cerr << "File utility.cpp: Velocity recording station switch parameter out of range!\n";
      std::exit(1);
    }
    if(problem_info.vel_global_end-problem_info.vel_global_start<0.)
    {
      std::cerr << "File utility.cpp: End of velocity recording station output comes before its begin!\n";
      std::exit(1);
    }
    param >> problem_info.hot_output >> problem_info.hot_freq;
    if(problem_info.hot_output<0 || problem_info.hot_output>2)
    {
      std::cerr << "File utility.cpp: Hot output parameter is out of range!\n";
      std::exit(1);
    }
    if(problem_info.hot_output>0 && problem_info.hot_freq<=0)
    {
      std::cerr << "File utility.cpp: Hot output frequency is out of range!\n";
      std::exit(1);
    }
    param.close();
  }
  catch(std::exception& ex)
  {
    std::cerr << "File utility.cpp: I/O error in configuration file" << ex.what() << '\n';
    std::exit(1);
  }
}

// Initialize degrees of freedom
void Initialize_DOF()
{
  for (int j=0; j<d_var; j++)
  {
//     int n_dim, order=problem_info.def_order-VAR_BASIS_INDEX[j];
    int n_dim, order=problem_info.var_order[j];
    std::size_t n_elem = VAR_2D_INDEX_INV[j]==BAD_INDEX ? 
      mesh.prism_list.size() : mesh.triangle_2d_list.size();

//     if(order<0)
//     {
//       std::cerr << "File utility.cpp: Negative alternative basis order!\n";
//       std::exit(1);
//     }
    if(VAR_2D_INDEX_INV[j]==BAD_INDEX)
      n_elem=mesh.prism_list.size(), n_dim=basis_ref[order]->Dim(); // 3D case
    else
      n_elem=mesh.triangle_2d_list.size(), n_dim=basis_2d[order]->Dim(); // 2D case
#ifdef BAROCLINIC_KEY
    // Special treatment for RHO_FRS variable
    if (j==RHO_FRS)
      n_dim=basis_2d[order]->Dim();
#endif

    var[j].InitializeGlobal(n_elem, n_dim);
    if (VAR_DT_INDEX_INV[j]!=BAD_INDEX)
    {
      for (int i=0; i<problem_info.runge_kutta_method-1; i++)
        var_stage[i][j].InitializeGlobal(n_elem, n_dim);
      var_rhs[j].InitializeGlobal(n_elem, n_dim);
      if (problem_info.diff_vert_switch && problem_info.diff_vert[j]!=0.)
        var_flux[j][2].InitializeGlobal(n_elem, n_dim);
      if (problem_info.diff_horiz_switch && problem_info.diff_horiz[j]!=0.)
        for (int i=0; i<d_space_2d; i++)
          var_flux[j][i].InitializeGlobal(n_elem, n_dim);
    }
  }
}
